from abc import ABC, abstractmethod
import subprocess
import scheduler
import surfex
import os
from datetime import datetime, timedelta
import signal
import time
import platform
import traceback
import sys
import shutil
import tomlkit
import json
# import math
from distutils.dir_util import copy_tree
sys.path.insert(0, "/usr/lib/python3/dist-packages/")
try:
    import ecflow
except ModuleNotFoundError:
    ecflow = None


# Base Scheduler server class
class Server(ABC):
    def __init__(self):
        pass

    @abstractmethod
    def create_suite(self, config, exp, suite_name, def_file, stream=None):
        raise NotImplementedError

    @abstractmethod
    def start_server(self):
        raise NotImplementedError

    @abstractmethod
    def replace(self, suite_name, def_file):
        raise NotImplementedError

    @abstractmethod
    def begin_suite(self, suite_name):
        raise NotImplementedError

    def start_exp(self, config, exp, suite_type, def_file, stream=None, begin=True):
        suite_name = self.create_suite(config, exp, suite_type, def_file, stream=stream)
        self.start_server()
        self.replace(suite_name, def_file)
        if begin:
            self.begin_suite(exp.name)


class EcflowServer(Server):
    def __init__(self, ecf_host, ecf_port, logfile, ):
        if ecflow is None:
            raise Exception("Ecflow was not found")
        Server.__init__(self)
        self.ecf_host = ecf_host
        self.ecf_port = ecf_port
        self.logfile = logfile
        self.ecf_client = ecflow.Client(self.ecf_host, self.ecf_port)

    def create_suite(self, config, exp, suite_type, def_file, stream=None):

        hh_list = exp.config.get_total_unique_hh_list()
        dtgstart = exp.progress.dtg
        dtgbeg = exp.progress.dtgbeg
        dtgend = exp.progress.dtgend
        print(dtgstart, dtgbeg, dtgend)
        if dtgbeg is None:
            dtgbeg = dtgstart
        if suite_type == "surfex":
            dtgs = []
            dtg = dtgstart
            while dtg <= dtgend:
                dtgs.append(dtg)
                hh = dtg.strftime("%H")
                fcint = None
                for h in range(0, len(hh_list)):
                    if int(hh_list[h]) == int(hh):
                        if h == len(hh_list) - 1:
                            fcint = ((int(hh_list[len(hh_list) - 1]) % 24) - int(hh_list[0])) % 24
                        else:
                            fcint = int(hh_list[h + 1]) - int(hh_list[h])
                if fcint is None:
                    raise Exception
                dtg = dtg + timedelta(hours=fcint)

            suite = scheduler.SurfexSuite(exp, dtgs, def_file, dtgbeg=dtgbeg)

        elif suite_type == "unittest":
            suite = scheduler.UnitTestSuite(exp, def_file)
        elif suite_type == "testbed":
            raise NotImplementedError
            # suite = scheduler.SurfexTestbedSuite(config, exp, def_file)
        else:
            raise Exception("Suite " + suite_type + " is not defined")

        if suite is not None:
            suite.save_as_defs()
        else:
            raise Exception("The suite is not constructed")
        return suite.suite_name

    def start_server(self):
        print("Start EcFlow server")
        try:
            self.ecf_client.ping()
            print("EcFlow server is already running")
        except RuntimeError:
            print("Re-Start EcFlow server")
            try:
                # Start server
                # self.ecf_client.restart_server()
                cmd = shutil.which("ecflow_start")
                if cmd is None:
                    cmd = shutil.which("ecflow_start.sh")
                if cmd is None:
                    raise Exception("ecflow_start not found")
                cmd = cmd + " -p " + str(self.ecf_port)
                print(cmd)
                ret = subprocess.call(cmd.split())
                if ret != 0:
                    raise RuntimeError
            except RuntimeError:
                raise Exception("Could not restart server!")

    def begin_suite(self, suite_name):
        self.ecf_client.begin_suite(suite_name)

    def force_complete(self, task):
        ecf_name = task.ecf_name
        self.ecf_client.force_state(ecf_name, ecflow.State.complete)

    def force_aborted(self, task):
        ecf_name = task.ecf_name
        self.ecf_client.force_state(ecf_name, ecflow.State.aborted)

    def update_submission_id(self, task):
        self.update_log(task.ecf_name)
        self.update_log(task.submission_id)
        print(task.ecf_name, "add", "variable", "SUBMISSION_ID", task.submission_id)
        self.ecf_client.alter(task.ecf_name, "add", "variable", "SUBMISSION_ID", task.submission_id)

    def replace(self, suite_name, def_file):
        print(suite_name, def_file)
        try:
            self.ecf_client.replace("/" + suite_name, def_file)
        except RuntimeError:
            try:
                self.ecf_client.delete("/" + suite_name)
                self.ecf_client.replace("/" + suite_name, def_file)
            except RuntimeError:
                raise Exception("Could not replace suite " + suite_name)

    def update_log(self, text):
        print(self.logfile)
        utcnow = datetime.utcnow().strftime("[%H:%M:%S %d.%m.%Y]")
        fh = open(self.logfile, "a")
        fh.write(utcnow + " " + str(text) + "\n")
        fh.flush()
        fh.close()


class EcflowServerFromFile(EcflowServer):
    def __init__(self, ecflow_server_file, logfile):
        if os.path.exists(ecflow_server_file):
            self.settings = surfex.toml_load(ecflow_server_file)
        else:
            raise FileNotFoundError("Could not find " + ecflow_server_file)

        ecf_host = self.get_var("ECF_HOST")
        ecf_port_offset = int(self.get_var("ECF_PORT_OFFSET", default=1500))
        ecf_port = int(self.get_var("ECF_PORT", default=int(os.getuid())))
        ecf_port = ecf_port + ecf_port_offset

        # logfile = self.get_var("SERVER_LOG")
        EcflowServer.__init__(self, ecf_host, ecf_port, logfile)

    def get_var(self, var, default=None):
        if var in self.settings:
            return self.settings[var]
        else:
            if default is not None:
                return default
            else:
                raise Exception("Variable " + var + " not found!")

    def save_as_file(self, wdir):
        fname = self.get_file_name(wdir)
        surfex.toml_dump(self.settings, fname)

    @staticmethod
    def get_file_name(wdir, full_path=False):
        f = "Env_server"
        if full_path:
            f = wdir + "/" + f
        return f


class EcflowLogServer(object):
    def __init__(self, ecf_loghost, ecf_logport):
        self.ecf_loghost = ecf_loghost
        self.ecf_logport = ecf_logport


class EcflowTask(object):
    def __init__(self, ecf_name, ecf_tryno, ecf_pass, ecf_rid, submission_id=None, ecf_timeout=20):
        self.ecf_name = ecf_name
        self.ecf_tryno = int(ecf_tryno)
        self.ecf_pass = ecf_pass
        if ecf_rid == "" or ecf_rid is None:
            ecf_rid = os.getpid()
        self.ecf_rid = int(ecf_rid)
        self.ecf_timeout = ecf_timeout
        ecf_name_parts = self.ecf_name.split("/")
        self.ecf_task = ecf_name_parts[-1]
        ecf_families = None
        if len(ecf_name_parts) > 2:
            ecf_families = ecf_name_parts[1:-1]
        self.ecf_families = ecf_families
        self.family1 = None
        if self.ecf_families is not None:
            self.family1 = self.ecf_families[-1]

        if submission_id == "":
            submission_id = None
        self.submission_id = submission_id

    def create_submission_log(self, joboutdir):
        return joboutdir + "/" + self.ecf_name + ".job" + str(self.ecf_tryno) + ".sub"

    def create_kill_log(self, joboutdir):
        return joboutdir + "/" + self.ecf_name + ".job" + str(self.ecf_tryno) + ".kill"

    def create_status_log(self, joboutdir):
        return joboutdir + "/" + self.ecf_name + ".job" + str(self.ecf_tryno) + ".stat"

    def create_ecf_job(self, joboutdir):
        return joboutdir + "/" + self.ecf_name + ".job" + str(self.ecf_tryno)

    def create_ecf_jobout(self, joboutdir):
        return joboutdir + "/" + self.ecf_name + "." + str(self.ecf_tryno)


class EcflowClient(object):
    """Encapsulate communication with the ecflow server. This will automatically call
       the child command init()/complete(), for job start/finish. It will also
       handle exceptions and signals, by calling the abort child command.
       *ONLY* one instance of this class, should be used. Otherwise zombies will be created.
    """

    def __init__(self, server, task):
        print("Creating Client")
        self.server = server
        self.ci = server.ecf_client
        # self.ci.set_host_port("%ECF_HOST%", "%ECF_PORT%")
        self.ci.set_child_pid(task.ecf_rid)
        self.ci.set_child_path(task.ecf_name)
        self.ci.set_child_password(task.ecf_pass)
        self.ci.set_child_try_no(task.ecf_tryno)
        print("   Only wait " + str(task.ecf_timeout) +
              " seconds, if the server cannot be contacted (note default is 24 hours) before failing")
        self.ci.set_child_timeout(task.ecf_timeout)
        # self.ci.set_zombie_child_timeout(10)
        self.task = task

        # Abort the task for the following signals
        signal.signal(signal.SIGINT, self.signal_handler)
        signal.signal(signal.SIGHUP, self.signal_handler)
        signal.signal(signal.SIGQUIT, self.signal_handler)
        signal.signal(signal.SIGILL, self.signal_handler)
        signal.signal(signal.SIGTRAP, self.signal_handler)
        signal.signal(signal.SIGIOT, self.signal_handler)
        signal.signal(signal.SIGBUS, self.signal_handler)
        signal.signal(signal.SIGFPE, self.signal_handler)
        signal.signal(signal.SIGUSR1, self.signal_handler)
        signal.signal(signal.SIGUSR2, self.signal_handler)
        signal.signal(signal.SIGPIPE, self.signal_handler)
        signal.signal(signal.SIGTERM, self.signal_handler)
        signal.signal(signal.SIGXCPU, self.signal_handler)
        if platform.system() != "Darwin":
            signal.signal(signal.SIGPWR, self.signal_handler)

    @staticmethod
    def at_time():
        return datetime.fromtimestamp(time.time()).strftime('%H:%M:%S')

    def signal_handler(self, signum, extra=None):
        print('   Aborting: Signal handler called with signal ', signum)
        # self.ci.child_abort("Signal handler called with signal " + str(signum))
        self.__exit__(Exception, "Signal handler called with signal " + str(signum), None)

    def __enter__(self):
        print('Calling init at: ' + self.at_time())
        # self.server.update_log(self.task.ecf_name + " init")
        self.ci.child_init()
        return self.ci

    def __exit__(self, ex_type, value, tb):
        print("   Client:__exit__: ex_type:" + str(ex_type) + " value:" + str(value))
        if ex_type is not None:
            print('Calling abort ' + self.at_time())
            self.ci.child_abort("Aborted with exception type " + str(ex_type) + ":" + str(value))
            if tb is not None:
                print(tb)
                traceback.print_tb(tb, limit=1, file=sys.stdout)
                print("*** print_exception:")
                # exc_type below is ignored on 3.5 and later
                print("*** print_exc:")
                traceback.print_exc(limit=2, file=sys.stdout)
                print("*** format_exc, first and last line:")
                formatted_lines = traceback.format_exc().splitlines()
                print(formatted_lines[0])
                print(formatted_lines[-1])
                print("*** format_exception:")
                print("*** extract_tb:")
                print(repr(traceback.extract_tb(tb)))
                print("*** format_tb:")
                print(repr(traceback.format_tb(tb)))
                print("*** tb_lineno:", tb.tb_lineno)
                self.server.update_log(self.task.ecf_name + " abort")
            return False
        print('Calling complete at: ' + self.at_time())
        # self.server.update_log(self.task.ecf_name + " complete")
        self.ci.child_complete()
        return False


class System(object):
    def __init__(self, host_system, exp_name):

        # print(host_system)
        self.system_variables = ["SFX_EXP_DATA", "SFX_EXP_LIB", "JOBOUTDIR", "MKDIR",
                                 "RSYNC", "HOST", "HOSTS", "HOST_NAME", "SURFEX_CONFIG"]
        self.hosts = None
        self.exp_name = exp_name

        # Set system0 from system_dict
        system0 = {}
        for var in self.system_variables:
            if var == "HOSTS":
                self.hosts = host_system["HOST_SYSTEM"]["HOSTS"]
            elif var == "HOST":
                pass
                ''' Now in Exp class
            elif var == "WD":
                self.wd = host_system["HOST_SYSTEM"]["WD"]
            elif var == "CONF":
                self.conf = host_system["HOST_SYSTEM"]["CONF"]
            elif var == "REV":
                self.rev = host_system["HOST_SYSTEM"]["REV"]
                '''
            else:
                if var in host_system["HOST_SYSTEM"]:
                    system0.update({var: host_system["HOST_SYSTEM"][var]})
                else:
                    raise Exception("Variable is missing: " + var)

        system = {}
        system.update({"HOSTS": self.hosts})
        for host in range(0, len(self.hosts)):
            systemn = system0.copy()
            systemn.update({"HOST": self.hosts[host]})
            hostn = "HOST" + str(host)
            if hostn in host_system["HOST_SYSTEM"]:
                for key in host_system["HOST_SYSTEM"][hostn]:
                    value = host_system["HOST_SYSTEM"][hostn][key]
                    # print(hostn, key, value)
                    systemn.update({key: value})
            system.update({str(host): systemn})

        self.system = system
        # Check for needed variables
        for var in self.system_variables:
            for host in range(0, len(self.hosts)):
                pass

    def get_var(self, var, host, stream=None):
        if var == "HOSTS":
            if self.hosts is not None:
                return self.hosts
            else:
                raise Exception("Variable " + var + " not found in system")
        else:
            os.environ.update({"EXP": self.exp_name})
            if var in self.system[str(host)]:
                if self.system[str(host)][var] is None:
                    raise Exception(var + " is None!")
                expanded_var = os.path.expandvars(self.system[str(host)][var])
                if stream is not None:
                    if var == "SFX_EXP_LIB":
                        expanded_var = expanded_var + stream
                return expanded_var
            else:
                raise Exception("Variable " + var + " not found in system")


class SystemFromFile(System):
    def __init__(self, env_system_file, exp_name):

        # system = System.get_file_name(wdir, full_path=True)
        print(env_system_file)
        if os.path.exists(env_system_file):
            host_system = surfex.toml_load(env_system_file)
        else:
            raise FileNotFoundError(env_system_file)

        # print(host_system)
        System.__init__(self, host_system, exp_name)


class Exp(object):
    def __init__(self, name, wdir, rev, conf, experiment_is_locked, system_file_paths=None, system=None, server=None,
                 configuration=None, geo=None, env_submit=None,  write_config_files=False, progress=None):

        self.name = name
        self.wd = wdir
        self.rev = rev
        self.conf = conf
        self.experiment_is_locked = experiment_is_locked
        self.system = system
        self.server = server
        self.env_submit = env_submit
        self.system_file_paths = system_file_paths
        self.progress = progress

        # Check existence of needed config files
        config = Exp.get_config(self.wd, self.conf)
        c_files = config["config_files"]
        config_files = {}
        for f in c_files:
            lfile = self.wd + "/config/" + f
            rfile = self.conf + "/scheduler/config/" + f

            if os.path.exists(lfile):
                # print("lfile", lfile)
                toml_dict = surfex.toml_load(lfile)
            else:
                if os.path.exists(rfile):
                    # print("rfile", rfile)
                    toml_dict = surfex.toml_load(rfile)
                else:
                    raise Exception("No config file found for " + f)

            config_files.update({
                f: {
                    "toml": toml_dict,
                    "blocks": config[f]["blocks"]
                }
            })
        self.config_files = config_files

        do_merge = False
        if configuration is not None:
            do_merge = True
            conf = self.wd + "/config/configurations/" + configuration.lower() + ".toml"
            if not os.path.exists(conf):
                conf = self.conf + "/scheduler/config/configurations/" + configuration.lower() + ".toml"
                if not os.path.exists(conf):
                    raise Exception("Can not find configuration " + configuration + " in: " + conf)
            configuration = surfex.toml_load(conf)

        if do_merge:
            self.merge_to_toml_config_files(configuration=configuration, write_config_files=write_config_files)

        # Merge config
        all_merged_settings = surfex.merge_toml_env_from_config_dicts(self.config_files)
        merged_config, member_merged_config = surfex.process_merged_settings(all_merged_settings)

        # Create configuration
        self.config = surfex.Configuration(merged_config, member_merged_config, geo=geo)

    def checkout(self, file):
        if file is None:
            raise Exception("File must be set")
        if os.path.exists(file):
            print("File is aleady checked out " + file)
        else:
            if os.path.exists(self.rev + "/" + file):
                dirname = os.path.dirname(self.wd + "/" + file)
                os.makedirs(dirname, exist_ok=True)
                shutil.copy2(self.rev + "/scheduler/" + file, self.wd + "/" + file)
                print("Checked out file: " + file)
            else:
                print("File was not found: " + self.rev + "/" + file)

    def setup_files(self, host):

        rev_file = Exp.get_file_name(self.wd, "rev", full_path=True)
        conf_file = Exp.get_file_name(self.wd, "conf", full_path=True)
        open(rev_file, "w").write(self.rev + "\n")
        open(conf_file, "w").write(self.conf + "\n")
        print("rev", self.rev, rev_file, "conf", self.conf, conf_file)

        env_system = Exp.get_file_name(self.wd, "system", full_path=False)
        env = Exp.get_file_name(self.wd, "env", full_path=False)
        env_submit = Exp.get_file_name(self.wd, "submit", full_path=False)
        env_server = Exp.get_file_name(self.wd, "server", full_path=False)
        input_paths = Exp.get_file_name(self.wd, "input_paths", full_path=False)
        # Create needed system files
        system_files = {
            env_system: "",
            env: "",
            env_submit: "",
            env_server: "",
            input_paths: "",
        }
        system_files.update({
            env_system: "config/system/" + host + ".toml",
            env: "config/env/" + host + ".py",
            env_submit: "config/submit/" + host + ".json",
            env_server: "config/server/" + host + ".toml",
            input_paths: "config/input_paths/" + host + ".json",
        })

        for key in system_files:

            target = self.wd + "/" + key
            lfile = self.wd + "/" + system_files[key]
            rfile = self.conf + "/scheduler/" + system_files[key]
            dirname = os.path.dirname(lfile)
            os.makedirs(dirname, exist_ok=True)
            if os.path.exists(lfile):
                print("System file " + lfile + " already exists, is not fetched again")
            else:
                shutil.copy2(rfile, lfile)
            if os.path.exists(target):
                print("System target file " + lfile + " already exists, is not fetched again")
            else:
                os.symlink(system_files[key], target)

        self.env_submit = json.load(open(self.wd + "/Env_submit", "r"))

        plib = self.wd + "/pysurfex"
        config_dirs = ["surfex", "scheduler", "bin"]
        for cdir in config_dirs:
            if not os.path.exists(plib + "/" + cdir):
                print("Copy " + cdir + " from " + self.conf)
                shutil.copytree(self.conf + "/" + cdir, plib + "/" + cdir,
                                ignore=shutil.ignore_patterns("config", "ecf", "nam", "toml"))
            else:
                print(cdir + " already exists in " + self.wd + "/pysurfex")

        # Check existence of needed config files
        local_config = self.wd + "/config/config.toml"
        rev_config = self.conf + "/scheduler/config/config.toml"
        config = surfex.toml_load(rev_config)
        c_files = config["config_files"]
        if os.path.exists(local_config):
            config = surfex.toml_load(local_config)
            c_files = config["config_files"]

        config_files = []
        for f in c_files:
            lfile = self.wd + "/config/" + f
            config_files.append(lfile)
            os.makedirs(self.wd + "/config", exist_ok=True)
            rfile = self.conf + "/scheduler/config/" + f
            if not os.path.exists(lfile):
                # print(rfile, lfile)
                shutil.copy2(rfile, lfile)
            else:
                print("Config file " + lfile + " already exists, is not fetched again")

        dirs = ["config/domains"]
        # Copy dirs
        for dir_path in dirs:
            os.makedirs(self.wd + "/" + dir_path, exist_ok=True)
            files = [f for f in os.listdir(self.conf + "/scheduler/" + dir_path)
                     if os.path.isfile(os.path.join(self.conf + "/scheduler/" + dir_path, f))]
            for f in files:
                print("f", f)
                fname = self.wd + "/" + dir_path + "/" + f
                rfname = self.conf + "/scheduler/" + dir_path + "/" + f
                if not os.path.exists(fname):
                    print("Copy " + rfname + " -> " + fname)
                    shutil.copy2(rfname, fname)

        # ECF_submit exceptions
        f = "config/submit/submission.json"
        fname = self.wd + "/" + f
        rfname = self.conf + "/scheduler/" + f
        if not os.path.exists(fname):
            print("Copy " + rfname + " -> " + fname)
            shutil.copy2(rfname, fname)

        # Init run
        files = ["ecf/InitRun.py", "ecf/default.py"]
        os.makedirs(self.wd + "/ecf", exist_ok=True)
        for f in files:
            fname = self.wd + "/" + f
            rfname = self.conf + "/scheduler/" + f
            if not os.path.exists(fname):
                print("Copy " + rfname + " -> " + fname)
                shutil.copy2(rfname, fname)

        exp_dirs = ["nam", "toml"]
        for exp_dir in exp_dirs:
            rdir = self.conf + "/scheduler/" + exp_dir
            ldir = self.wd + "/" + exp_dir
            print("Copy " + rdir + " -> " + ldir)
            # shutil.copytree(rdir, ldir)
            copy_tree(rdir, ldir)

    def merge_testbed_submit(self, testbed_submit, decomposition="2D"):
        if os.path.exists(testbed_submit):
            testbed_submit = json.load(open(testbed_submit, "r"))
        if decomposition not in testbed_submit:
            raise Exception("Decomposition " + decomposition + " not found in " + testbed_submit)
        return surfex.merge_toml_env(self.env_submit, testbed_submit[decomposition])

    def merge_testbed_configurations(self, testbed_confs):
        merged_conf = {}

        for testbed_configuration in testbed_confs:
            print("Merging testbed configuration: " + testbed_configuration)
            testbed_configuration = "/config/testbed/" + testbed_configuration
            if os.path.exists(self.wd + testbed_configuration):
                testbed_configuration = self.wd + testbed_configuration
            else:
                if os.path.exists(self.rev + testbed_configuration):
                    testbed_configuration = self.rev + testbed_configuration
                else:
                    raise Exception("Testbed configuration not existing: " + testbed_configuration)

            conf = surfex.toml_load(testbed_configuration)
            merged_conf = surfex.merge_toml_env(merged_conf, conf)

        return merged_conf

    def merge_to_toml_config_files(self, configuration=None, testbed_configuration=None,
                                   user_settings=None,
                                   write_config_files=True):

        self.config_files = surfex.merge_config_files_dict(self.config_files, configuration=configuration,
                                                           testbed_configuration=testbed_configuration,
                                                           user_settings=user_settings)

        for f in self.config_files:
            this_config_file = "config/" + f

            block_config = self.config_files[f]["toml"]
            if write_config_files:
                f_out = self.wd + "/" + this_config_file
                dirname = os.path.dirname(f_out)
                # print(dirname)
                dirs = dirname.split("/")
                # print(dirs)
                if len(dirs) > 1:
                    p = "/"
                    for d in dirs[1:]:
                        p = p + d
                        # print(p)
                        os.makedirs(p, exist_ok=True)
                        p = p + "/"
                f_out = open(f_out, "w")
                f_out.write(tomlkit.dumps(block_config))

    @staticmethod
    def get_file_name(wd, ftype, full_path=False, stream=None):
        if ftype == "rev" or ftype == "conf":
            f = ftype
        elif ftype == "submit":
            f = "Env_submit"
        elif ftype == "system":
            f = "Env_system"
        elif ftype == "env":
            f = "Env"
        elif ftype == "server":
            f = "Env_server"
        elif ftype == "input_paths":
            f = "Env_input_paths"
        elif ftype == "progress":
            f = "progress.toml"
            if stream is not None:
                f = "progress" + stream + ".toml"
        elif ftype == "progressPP":
            f = "progressPP.toml"
            if stream is not None:
                f = "progressPP" + stream + ".toml"
        elif ftype == "domain":
            f = "domain.json"
        else:
            raise Exception
        if full_path:
            return wd + "/" + f
        else:
            return f

    @staticmethod
    def get_config(wdir, conf):
        # Check existence of needed config files
        local_config = wdir + "/config/config.toml"
        rev_config = conf + "/scheduler/config/config.toml"
        if os.path.exists(local_config):
            c_files = surfex.toml_load(local_config)
        elif os.path.exists(rev_config):
            c_files = surfex.toml_load(rev_config)
        else:
            raise Exception("No config found in " + wdir + " or " + conf)
        return c_files

    @staticmethod
    def get_experiment_is_locked_file(wdir, stream=None, full_path=True):

        experiment_is_locked_file = "experiment_is_locked"
        if stream is not None:
            experiment_is_locked_file = experiment_is_locked_file + stream

        if full_path:
            experiment_is_locked_file = wdir + "/" + experiment_is_locked_file
        return experiment_is_locked_file


class ExpFromFiles(Exp):
    def __init__(self, name, wdir, stream=None, host="0", progress=None):

        rev_file = Exp.get_file_name(wdir, "rev", full_path=True)
        conf_file = Exp.get_file_name(wdir, "conf", full_path=True)
        env_submit_file = Exp.get_file_name(wdir, "submit", full_path=True)

        # print(rev_file)
        if os.path.exists(rev_file):
            rev = open(rev_file, "r").read().rstrip()
        else:
            raise FileNotFoundError(rev_file)

        if os.path.exists(conf_file):
            conf = open(conf_file, "r").read().rstrip()
        else:
            raise FileNotFoundError(rev_file)

        # Check existence of needed system files
        system_files = {
            Exp.get_file_name(wdir, "system"): "",
            Exp.get_file_name(wdir, "env"): "",
            Exp.get_file_name(wdir, "submit"): ""
        }

        # Check for needed system files
        for key in system_files:
            target = wdir + "/" + key
            if not os.path.exists(target):
                raise Exception("System target file is missing " + target)

        c_files = Exp.get_config(wdir, conf)["config_files"]
        config_files = []
        for f in c_files:
            lfile = wdir + "/config/" + f
            config_files.append(lfile)
            if not os.path.exists(lfile):
                raise Exception("Needed config file is missing " + f)

        experiment_is_locked_file = Exp.get_experiment_is_locked_file(wdir, stream=stream, full_path=True)
        if os.path.exists(experiment_is_locked_file):
            experiment_is_locked = True
        else:
            experiment_is_locked = False

        # System
        env_system = Exp.get_file_name(wdir, "system", full_path=True)
        system = SystemFromFile(env_system, name)

        # System file path
        input_paths = self.get_file_name(wdir, "input_paths", full_path=True)
        if os.path.exists(input_paths):
            system_file_paths = json.load(open(input_paths, "r"))
        else:
            raise FileNotFoundError("System setting input paths not found " + input_paths)
        system_file_paths = SystemFilePathsFromSystem(system_file_paths, system, host=host, stream=stream)

        env_submit = json.load(open(env_submit_file, "r"))

        logfile = system.get_var("SFX_EXP_DATA", "0") + "/ECF.log"
        server = scheduler.EcflowServerFromFile(Exp.get_file_name(wdir, "server", full_path=True), logfile)

        domain_file = self.get_file_name(wdir, "domain", full_path=True)
        geo = surfex.geo.get_geo_object(json.load(open(domain_file, "r")))

        print("progress")
        if progress is None:
            progress = ProgressFromFile(self.get_file_name(wdir, "progress", full_path=True),
                                        self.get_file_name(wdir, "progressPP", full_path=True))

        Exp.__init__(self, name, wdir, rev, conf, experiment_is_locked, system_file_paths=system_file_paths,
                     system=system, server=server, env_submit=env_submit, geo=geo, progress=progress)

    def set_experiment_is_locked(self, stream=None):
        experiment_is_locked_file = Exp.get_experiment_is_locked_file(self.wd, stream=stream, full_path=True)
        fh = open(experiment_is_locked_file, "w")
        fh.write("Something from git?")
        fh.close()
        self.experiment_is_locked = True


class Progress(object):
    def __init__(self, progress, progress_pp):

        # Update DTG/DTGBED/DTGEND
        if "DTG" in progress:
            dtg = progress["DTG"]
            # Dump DTG to progress
            if "DTGEND" in progress:
                dtgend = progress["DTGEND"]
            else:
                if "DTGEND" in progress:
                    dtgend = progress["DTGEND"]
                else:
                    dtgend = progress["DTG"]

            if "DTGBEG" in progress:
                dtgbeg = progress["DTGBEG"]
            else:
                if "DTG" in progress:
                    dtgbeg = progress["DTG"]
                else:
                    raise Exception("Can not set DTGBEG")
            if dtgbeg is not None:
                if isinstance(dtgbeg, str):
                    dtgbeg = datetime.strptime(dtgbeg, "%Y%m%d%H")
                self.dtgbeg = dtgbeg
            else:
                self.dtgbeg = None
            if dtg is not None:
                if isinstance(dtg, str):
                    dtg = datetime.strptime(dtg, "%Y%m%d%H")
                self.dtg = dtg
            else:
                self.dtg = None
            if dtgend is not None:
                if isinstance(dtgend, str):
                    dtgend = datetime.strptime(dtgend, "%Y%m%d%H")
                self.dtgend = dtgend
            else:
                self.dtgend = None
        else:
            raise Exception

        # Update DTGPP
        dtgpp = None
        if "DTGPP" in progress_pp:
            dtgpp = progress_pp["DTGPP"]
        elif "DTG" in progress:
            dtgpp = progress["DTG"]
        if dtgpp is not None:
            if isinstance(dtgpp, str):
                dtgpp = datetime.strptime(dtgpp, "%Y%m%d%H")
            self.dtgpp = dtgpp

        print("DTGEND", self.dtgend)

    def export_to_file(self, fname):
        fh = open(fname, "w")
        fh.write("export DTG=" + self.dtg.strftime("%Y%m%d%H") + "\n")
        fh.write("export DTGBEG=" + self.dtgbeg.strftime("%Y%m%d%H") + "\n")
        fh.write("export DTGEND=" + self.dtgend.strftime("%Y%m%d%H") + "\n")
        fh.write("export DTGPP=" + self.dtgpp.strftime("%Y%m%d%H") + "\n")
        fh.close()

    # Members could potentially have different DTGBEGs
    def get_dtgbeg(self, fcint):
        dtgbeg = self.dtgbeg
        if (self.dtg - timedelta(hours=int(fcint))) < self.dtgbeg:
            dtgbeg = self.dtg
        return dtgbeg

    # Members could potentially have different DTGENDs
    def get_dtgend(self, fcint):
        dtgend = self.dtgend
        if self.dtgend < (self.dtg + timedelta(hours=int(fcint))):
            dtgend = self.dtg
        return dtgend

    def increment_progress(self, fcint_min, pp=False):
        if pp:
            self.dtgpp = self.dtgpp + timedelta(hours=fcint_min)
        else:
            self.dtg = self.dtg + timedelta(hours=fcint_min)
            if self.dtgend < self.dtg:
                self.dtgend = self.dtg

    def save(self, progress_file, progress_pp_file, log=True, log_pp=True):
        progress = {
            "DTGBEG": self.dtgbeg.strftime("%Y%m%d%H"),
            "DTG": self.dtg.strftime("%Y%m%d%H"),
            "DTGEND": self.dtgend.strftime("%Y%m%d%H"),
        }
        progress_pp = {
            "DTGPP": self.dtgpp.strftime("%Y%m%d%H"),
        }
        if log:
            surfex.toml_dump(progress, progress_file)
        if log_pp:
            surfex.toml_dump(progress_pp, progress_pp_file)


class ProgressFromFile(Progress):
    def __init__(self, progress, progress_pp):

        self.progress_file = progress
        self.progress_pp_file = progress_pp
        if os.path.exists(self.progress_file):
            progress = surfex.toml_load(self.progress_file)
        else:
            progress = {
                "DTGBEG": None,
                "DTG": None,
                "DTGEND": None
            }
        if os.path.exists(self.progress_pp_file):
            progress_pp = surfex.toml_load(self.progress_pp_file)
        else:
            progress_pp = {
                "DTGPP": None
            }

        Progress.__init__(self, progress, progress_pp)

    def increment_progress(self, fcint_min, pp=False):
        Progress.increment_progress(self, fcint_min, pp=False)
        if pp:
            updated_progress_pp = {
                "DTGPP": self.dtgpp.strftime("%Y%m%d%H")
            }
            surfex.toml_dump(updated_progress_pp, self.progress_pp_file)
        else:
            updated_progress = {
                "DTGBEG": self.dtgbeg.strftime("%Y%m%d%H"),
                "DTG": self.dtg.strftime("%Y%m%d%H"),
                "DTGEND": self.dtgend.strftime("%Y%m%d%H")
            }
            surfex.toml_dump(updated_progress, self.progress_file)


class SystemFilePathsFromSystem(surfex.SystemFilePaths):

    """

    Also set SFX_EXP system variables (File stucture/ssh etc)

    """
    def __init__(self, paths, system, **kwargs):
        surfex.SystemFilePaths.__init__(self, paths)
        host = "0"
        if "host" in kwargs:
            host = kwargs["host"]
        stream = None
        if "stream" in kwargs:
            stream = kwargs["stream"]
        verbosity = 0

        # override paths from system file
        sfx_data = self.substitute_string(system.get_var("SFX_EXP_DATA", host=host, stream=stream))
        sfx_lib = self.substitute_string(system.get_var("SFX_EXP_LIB", host=host, stream=stream))
        self.sfx_exp_vars = {
            "SFX_EXP_DATA": sfx_data,
            "SFX_EXP_LIB": sfx_lib
        }

        os.makedirs(sfx_data, exist_ok=True)
        os.makedirs(sfx_lib, exist_ok=True)
        self.system_variables = {"SFX_EXP_DATA": sfx_data, "SFX_EXP_LIB": sfx_lib}

        self.add_system_file_path("sfx_exp_data", sfx_data)
        self.add_system_file_path("sfx_exp_lib", sfx_lib)
        self.add_system_file_path("default_archive_dir", sfx_data + "/archive/@YYYY@/@MM@/@DD@/@HH@/@EEE@/",
                                  check_parsing=False)
        archive_dir = self.get_system_path("archive_dir",
                                           default_dir="default_archive_dir", verbosity=verbosity, check_parsing=False)
        self.add_system_file_path("archive_dir", archive_dir, check_parsing=False)
        self.sfx_exp_vars.update({"ARCHIVE": archive_dir})

        self.add_system_file_path("extrarch_dir", sfx_data + "/archive/extract/@EEE@/", check_parsing=False)
        self.add_system_file_path("climdir", sfx_data + "/climate/@EEE@/", check_parsing=False)
        bindir = self.get_system_path("bin_dir", default_dir=sfx_data + "/bin/")
        self.add_system_file_path("bin_dir", bindir)
        self.add_system_file_path("wrk_dir", sfx_data + "/@YYYY@@MM@@DD@_@HH@/@EEE@/", check_parsing=False)
        self.add_system_file_path("forcing_dir", sfx_data + "/forcing/@YYYY@@MM@@DD@@HH@/@EEE@/", check_parsing=False)

        climdir = self.get_system_path("climdir", check_parsing=False)
        self.add_system_file_path("pgd_dir", climdir, check_parsing=False)
        archive = self.get_system_path("archive_dir", check_parsing=False)
        self.add_system_file_path("prep_dir", archive, check_parsing=False)
        self.add_system_file_path("default_obs_dir", sfx_data + "/archive/observations/@YYYY@/@MM@/@DD@/@HH@/@EEE@/",
                                  check_parsing=False)
        obs_dir = self.get_system_path("obs_dir", default_dir="default_obs_dir", verbosity=verbosity,
                                       check_parsing=False)
        self.sfx_exp_vars.update({"OBDIR": obs_dir})
        self.add_system_file_path("obs_dir", obs_dir)
        first_guess_dir = self.get_system_path("archive_dir", check_parsing=False)
        self.add_system_file_path("first_guess_dir", first_guess_dir, check_parsing=False)

    def get_system_path(self, dtype, **kwargs):
        kwargs.update({
            "sfx_exp_vars": self.sfx_exp_vars
        })
        return surfex.SystemFilePaths.get_system_path(self,  dtype, **kwargs)


class SystemFilePathsFromSystemFile(SystemFilePathsFromSystem):

    """

    Also set SFX_EXP system variables (File stucture/ssh etc)

    """
    def __init__(self, system_file_paths, system, name, **kwargs):
        system_file_paths = json.load(open(system_file_paths, "r"))
        system = SystemFromFile(system, name)
        SystemFilePathsFromSystem.__init__(self, system_file_paths, system, **kwargs)


def init_run(exp, stream=None):

    system = exp.system
    hosts = exp.system.hosts
    wd = exp.wd

    rsync = str(system.get_var("RSYNC", "0", stream=stream))
    lib0 = system.get_var("SFX_EXP_LIB", "0", stream=stream)
    rev = exp.rev
    host_name0 = system.get_var("HOST_NAME", "0", stream=stream)
    if host_name0 != "":
        host_name0 = os.environ["USER"] + "@" + host_name0 + ":"

    # Sync CONF to LIB0
    if not exp.experiment_is_locked:
        os.makedirs(lib0 + "/pysurfex", exist_ok=True)
        dirs = ["surfex", "scheduler", "bin", "test"]
        for d in dirs:
            cmd = rsync + " " + exp.conf + "/" + d + "/ " + host_name0 + lib0 + "/pysurfex/" + d + \
                  " --exclude=.git --exclude=nam --exclude=toml --exclude=config --exclude=ecf " + \
                  "--exclude=__pycache__ --exclude='*.pyc'"
            print(cmd)
            ret = subprocess.call(cmd, shell=True)
            if ret != 0:
                raise Exception

        dirs = ["nam", "toml", "config", "ecf"]
        for d in dirs:
            cmd = rsync + " " + exp.conf + "/scheduler/" + d + "/ " + host_name0 + lib0 + "/" + d + \
                  " --exclude=.git --exclude=__pycache__ --exclude='*.pyc'"
            print(cmd)
            ret = subprocess.call(cmd, shell=True)
            if ret != 0:
                raise Exception
    else:
        print("Not resyncing CONF as experiment is locked")

    # Sync REV to LIB0
    if not exp.experiment_is_locked:
        if rev != wd:
            # print(host_name0)
            # print(lib0)
            cmd = rsync + " " + rev + "/ " + host_name0 + lib0 + " --exclude=.git"
            print(cmd)
            ret = subprocess.call(cmd, shell=True)
            if ret != 0:
                raise Exception
        else:
            print("REV == WD. No syncing needed")
    else:
        print("Not resyncing REV as experiment is locked")

    # Sync WD to LIB
    # Always sync WD unless it is not same as SFX_EXP_LIB
    if wd != lib0:
        cmd = rsync + " " + wd + "/ " + host_name0 + lib0 + " --exclude=.git"
        print(cmd)
        ret = subprocess.call(cmd, shell=True)
        if ret != 0:
            raise Exception

    # Sync HM_LIB beween hosts
    if len(hosts) > 1:
        for host in range(1, len(hosts)):
            host = str(host)
            print("Syncing to HOST" + host)
            libn = system.get_var("SFX_EXP_LIB", host, stream=stream)
            datan = system.get_var("SFX_EXP_DATA", host, stream=stream)
            mkdirn = system.get_var("MKDIR", host, stream=stream)
            host_namen = system.get_var("HOST_NAME", host, stream=stream)
            ssh = ""
            if host_namen != "":
                ssh = "ssh " + os.environ["USER"] + "@" + host_namen
                host_namen = os.environ["USER"] + "@" + host_namen + ":"

            cmd = mkdirn + " " + datan
            print(cmd)
            ret = subprocess.call(cmd, shell=True)
            if ret != 0:
                raise Exception
            cmd = mkdirn + " " + libn
            if ssh != "":
                cmd = ssh + " \"" + mkdirn + " " + libn + "\""
            print(cmd)
            subprocess.call(cmd, shell=True)
            if ret != 0:
                raise Exception
            cmd = rsync + " " + host_name0 + lib0 + "/ " + host_namen + libn + " --exclude=.git"
            print(cmd)
            subprocess.call(cmd, shell=True)
            if ret != 0:
                raise Exception

    print("Lock experiment")
    exp.set_experiment_is_locked(stream=stream)
    print("Finished syncing")



import unittest
import scheduler
import surfex
import os
import json
from datetime import datetime


class TestSubmit(unittest.TestCase):

    @staticmethod
    def _create_exp(wd):
        os.makedirs(wd, exist_ok=True)
        argv = [
            "setup",
            "--wd", wd,
            "-rev", wd,
            "-conf", os.getcwd(),
            "-host", "unittest",
            "--domain_file", "test/settings/conf_proj_test.json"
        ]
        kwargs = scheduler.parse_surfex_script(argv)
        scheduler.surfex_script(**kwargs)

    def test_submission(self):

        exp = "test_submission"
        hm_wd = "/tmp/host0/hm_wd/" + exp
        self._create_exp(hm_wd)

        # Set progress
        dtg = datetime(2020, 6, 9, 9)
        progress = {"DTG": dtg.strftime("%Y%m%d%H")}
        progress_pp = {"DTGPP": dtg.strftime("%Y%m%d%H")}
        progress = scheduler.Progress(progress, progress_pp)

        sfx_exp = scheduler.ExpFromFiles(exp, hm_wd, progress=progress)

        ecf_name = "/" + exp + "/Forecasting/Forecast"
        ecf_tryno = 1
        ecf_pass = "dummy_password"

        task = scheduler.EcflowTask(ecf_name, ecf_tryno, ecf_pass, ecf_rid=int(os.getpid()),
                                    submission_id=None)
        os.makedirs("/tmp/host1/job/" + exp + "/Forecasting/", exist_ok=True)
        fh = open("/tmp/host1/job/" + exp + "/Forecasting/Forecast.job1", "w")
        fh.write("My job generated by ecflow")
        fh.close()

        exceptions = {
            "complete": {
                "task": {
                    "Forecast": "is_coldstart"
                },
                "family": {
                    "Forecasting": "is_coldstart"
                }
            }
        }

        env_file = "/tmp/host1/job/" + exp + "/Forecasting/Env"
        fh = open(env_file, "w")
        fh.write("print(\"Oh my environment\n\")")
        fh.close()

        hosts = sfx_exp.system.hosts
        joboutdir = {}
        for host in range(0, len(hosts)):
            joboutdir.update({str(host): sfx_exp.system.get_var("JOBOUTDIR", str(host))})
        submit = scheduler.EcflowSubmitTask(task, sfx_exp.env_submit, sfx_exp.server, joboutdir,
                                            submit_exceptions=exceptions, env_file=env_file)
        submit.write_job()

    def test_slurm(self):

        exp = "test_slurm"

        ecf_name = "/" + exp + "/Forecasting/Forecast"
        ecf_tryno = 1
        ecf_pass = "dummy_password"

        task = scheduler.EcflowTask(ecf_name, ecf_tryno, ecf_pass, ecf_rid=int(os.getpid()),
                                    submission_id=None)
        os.makedirs("/tmp/host1/job/" + exp + "/Forecasting/", exist_ok=True)
        fh = open("/tmp/host1/job/" + exp + "/Forecasting/Forecast.job1", "w")
        fh.write("My job generated by ecflow")
        fh.close()

        joboutdirs = {
            "0": "/tmp/host0/job/",
            "1": "/tmp/host1/job/"
        }

        env_submit = {
            "submit_types": ["background", "scalar"],
            "default_submit_type": "scalar",
            "background": {
                "HOST": "0",
                "OMP_NUM_THREADS": "import os\nos.environ.update({\"OMP_NUM_THREADS\": \"1\"})",
                "tasks": [
                    "InitRun",
                    "LogProgress",
                    "LogProgressPP"
                ]
            },
            "scalar": {
                "HOST": "1",
                "SUBMIT_TYPE": "slurm",
                "SSH": "ssh " + os.environ["USER"] + "@localhost",
                "INTERPRETER": "#!/usr/bin/env python3",
                "WRAPPER": "wrapper",
                "Not_existing_task": {
                    "DR_HOOK": "print(\"hei\")"
                }
            }
        }

        task_settings = scheduler.TaskSettings(task, env_submit, joboutdirs)
        sub = scheduler.SlurmSubmission(task, task_settings, sub=os.getcwd() + "/test/bin/sbatch")
        sub.set_submit_cmd()
        sub.submit_job()
        sub.set_jobid()
        self.assertEqual(sub.job_id, "slurm.12345")
        if not os.path.exists("/tmp/host1/job/" + exp + "/Forecasting/Forecast.job1.sub"):
            raise Exception("Expected sub file mot found")

        task = scheduler.EcflowTask(ecf_name, ecf_tryno, ecf_pass, ecf_rid=int(os.getpid()),
                                    submission_id=sub.job_id)
        sub = scheduler.get_submission_object(task, task_settings)
        sub.status()
        if not os.path.exists("/tmp/host1/job/" + exp + "/Forecasting/Forecast.job1.stat"):
            raise Exception("Expected status file mot found")
        sub.kill()
        if not os.path.exists("/tmp/host1/job/" + exp + "/Forecasting/Forecast.job1.kill"):
            raise Exception("Expected kill file mot found")

    def test_pbs(self):
        exp = "test_pbs"

        ecf_name = "/" + exp + "/Forecasting/Forecast"
        ecf_tryno = 1
        ecf_pass = "dummy_password"

        task = scheduler.EcflowTask(ecf_name, ecf_tryno, ecf_pass, ecf_rid=int(os.getpid()),
                                    submission_id=None)
        os.makedirs("/tmp/host1/job/" + exp + "/Forecasting/", exist_ok=True)
        fh = open("/tmp/host1/job/" + exp + "/Forecasting/Forecast.job1", "w")
        fh.write("My job generated by ecflow")
        fh.close()

        joboutdirs = {
            "0": "/tmp/host0/job/",
            "1": "/tmp/host1/job/"
        }

        env_submit = {
            "submit_types": ["background", "scalar"],
            "default_submit_type": "scalar",
            "background": {
                "HOST": "0",
                "OMP_NUM_THREADS": "import os\nos.environ.update({\"OMP_NUM_THREADS\": \"1\"})",
                "tasks": [
                    "InitRun",
                    "LogProgress",
                    "LogProgressPP"
                ]
            },
            "scalar": {
                "HOST": "1",
                "SUBMIT_TYPE": "pbs",
                "Not_existing_task": {
                    "DR_HOOK": "print(\"hei\")"
                }
            }
        }

        task_settings = scheduler.TaskSettings(task, env_submit, joboutdirs)
        sub = scheduler.get_submission_object(task, task_settings)
        sub.set_submit_cmd()
        sub.submit_job()
        sub.set_jobid()
        self.assertEqual(sub.job_id, "pbs.12345")
        if not os.path.exists("/tmp/host1/job/" + exp + "/Forecasting/Forecast.job1.sub"):
            raise Exception("Expected sub file mot found")

        task = scheduler.EcflowTask(ecf_name, ecf_tryno, ecf_pass, ecf_rid=int(os.getpid()),
                                    submission_id=sub.job_id)
        sub = scheduler.PBSSubmission(task, task_settings)
        sub.status()
        if not os.path.exists("/tmp/host1/job/" + exp + "/Forecasting/Forecast.job1.stat"):
            raise Exception("Expected status file mot found")
        sub.kill()
        if not os.path.exists("/tmp/host1/job/" + exp + "/Forecasting/Forecast.job1.kill"):
            raise Exception("Expected kill file mot found")

    def test_grid_engine(self):
        exp = "test_grid_engine"

        ecf_name = "/" + exp + "/Forecasting/Forecast"
        ecf_tryno = 1
        ecf_pass = "dummy_password"

        task = scheduler.EcflowTask(ecf_name, ecf_tryno, ecf_pass, ecf_rid=int(os.getpid()),
                                    submission_id=None)
        os.makedirs("/tmp/host1/job/" + exp + "/Forecasting/", exist_ok=True)
        fh = open("/tmp/host1/job/" + exp + "/Forecasting/Forecast.job1", "w")
        fh.write("My job generated by ecflow")
        fh.close()

        joboutdirs = {
            "0": "/tmp/host0/job/",
            "1": "/tmp/host1/job/"
        }

        env_submit = {
            "submit_types": ["background", "scalar"],
            "default_submit_type": "scalar",
            "background": {
                "HOST": "0",
                "OMP_NUM_THREADS": "import os\nos.environ.update({\"OMP_NUM_THREADS\": \"1\"})",
                "tasks": [
                    "InitRun",
                    "LogProgress",
                    "LogProgressPP"
                ]
            },
            "scalar": {
                "HOST": "1",
                "SUBMIT_TYPE": "grid_engine",
                "SUBMIT": "qsub_sge",
                "Not_existing_task": {
                    "DR_HOOK": "print(\"hei\")"
                }
            }
        }

        task_settings = scheduler.TaskSettings(task, env_submit, joboutdirs)
        # Test default
        scheduler.get_submission_object(task, task_settings)
        # Make my own test SGE
        sub = scheduler.GridEngineSubmission(task, task_settings, sub="qsub_sge", stat="qacct_sge -j", kill="qdel_sge")
        sub.set_submit_cmd()
        sub.submit_job()
        sub.set_jobid()
        self.assertEqual(sub.job_id, "sge.12345")
        if not os.path.exists("/tmp/host1/job/" + exp + "/Forecasting/Forecast.job1.sub"):
            raise Exception("Expected sub file mot found")

        task = scheduler.EcflowTask(ecf_name, ecf_tryno, ecf_pass, ecf_rid=int(os.getpid()),
                                    submission_id=sub.job_id)
        sub = scheduler.GridEngineSubmission(task, task_settings, sub="qsub_sge", stat="qacct_sge -j", kill="qdel_sge")
        sub.status()
        if not os.path.exists("/tmp/host1/job/" + exp + "/Forecasting/Forecast.job1.stat"):
            raise Exception("Expected status file mot found")
        sub.kill()
        if not os.path.exists("/tmp/host1/job/" + exp + "/Forecasting/Forecast.job1.kill"):
            raise Exception("Expected kill file mot found")



import shutil
import os
import subprocess
from abc import ABC, abstractmethod


class EcflowSubmitTask(object):
    def __init__(self, task, env_submit, server, joboutdir,
                 stream=None, dbfile=None, interpreter="#!/usr/bin/env python3",
                 ensmbr=None, submit_exceptions=None, coldstart=False, env_file=None, communication=True):
        self.task = task
        self.env_file = env_file
        self.ecflow_server = server
        self.coldstart = coldstart
        self.communication = communication

        self.ensmbr = ensmbr
        if self.ensmbr is not None:
            if ensmbr < 0:
                self.ensmbr = None

        self.db_file = dbfile
        self.stream = stream
        self.complete = False

        # Parse Env_submit
        self.task_settings = TaskSettings(self.task, env_submit, joboutdir, interpreter=interpreter,
                                          submit_exceptions=submit_exceptions, coldstart=False)
        self.sub = get_submission_object(self.task, self.task_settings, db_file=self.db_file)

    def write_header(self, fh):
        fh.write(self.task_settings.interpreter + "\n")
        if self.task_settings.header is not None:
            fh.write("\n# Batch commands\n")
            # Loop twice, first comments (likely to be batch commands)
            for setting in self.task_settings.header:
                value = str(self.task_settings.header[setting])
                if value.find("#") >= 0:
                    # print(str(self.header[setting]))
                    fh.write(str(self.task_settings.header[setting]) + "\n")
            
            # Host environment
            if self.env_file is not None:
                fh.write("\n# Host specific environment settings in python syntax:\n")
                fh_env = open(self.env_file, "r")
                for line in fh_env.readlines():
                    fh.write(line)
                fh_env.close()

            fh.write("\n# Task specific settings:\n")
            for setting in self.task_settings.header:
                value = str(self.task_settings.header[setting])
                if value.find("#") < 0:
                    # print(str(self.header[setting]))
                    fh.write(str(self.task_settings.header[setting]) + "\n")
             
            wrapper = ""
            if self.task_settings.wrapper is not None:
                wrapper = str(self.task_settings.wrapper)

            fh.write("\n#Python script:\n")
            return wrapper, self.task_settings.host

    def write_trailer(self, fh):
        if self.task_settings.trailer is not None:
            for setting in self.task_settings.trailer:
                fh.write(str(self.task_settings.trailer[setting]) + "\n")

    def write_job(self):
        ecf_job = self.task_settings.ecf_job_at_host
        fname = ecf_job + ".tmp"
        shutil.move(ecf_job, fname)
        fh = open(ecf_job, "w")
        wrapper, host = self.write_header(fh)
        job_fh = open(fname, "r")
        for line in job_fh.readlines():
            line = line.replace("@WRAPPER_TO_BE_SUBSTITUTED@", wrapper)
            line = line.replace("@HOST_TO_BE_SUBSTITUTED@", host)
            fh.write(line)
        job_fh.close()
        self.write_trailer(fh)
        fh.close()
        os.system("chmod u+x " + ecf_job)

    def submit(self):
        try:
            if "OUTPUT" not in self.task_settings.header:
                self.task_settings.header.update({"OUTPUT": self.sub.set_output()})
            if "NAME" not in self.task_settings.header:
                self.task_settings.header.update({"NAME": self.sub.set_job_name()})
            self.write_job()

            if self.complete:
                self.ecflow_server.force_complete(self.task)
            else:
                self.sub.set_submit_cmd()
                self.sub.submit_job()
                self.sub.set_jobid()
                self.task.submission_id = self.sub.job_id
                self.ecflow_server.update_submission_id(self.task)

        except RuntimeError:
            # Supposed to handle abort it self unless killed
            pass
        except Exception as error:
            raise SubmitException("Submission failed " + repr(error), self.task, self.task_settings)


class TaskSettings(object):

    def __init__(self, task, submission_defs, joboutdirs, submit_exceptions=None, interpreter="#!/usr/bin/env python3",
                 complete=False, coldstart=False):
        self.task = task
        self.submission_defs = submission_defs
        self.header = {}
        self.trailer = {}
        self.wrapper = None
        self.submit_type = "background"
        self.interpreter = interpreter
        self.complete = complete
        self.remote_submit_cmd = None
        self.remote_status_cmd = None
        self.remote_kill_cmd = None
        self.coldstart = coldstart
        self.host = None
        if submit_exceptions is not None:
            self.check_exceptions(submit_exceptions)
        self.task_settings = self.parse_submission_defs()
        self.process_settings()

        if self.host is None:
            raise Exception("Host number is mandatory!")

        if "0" in joboutdirs:
            joboutdir = joboutdirs["0"]
        else:
            raise Exception("No joboutdir defined for HOST 0")

        print("Task HOST is ", self.host)
        if self.host in joboutdirs:
            joboutdir_at_host = joboutdirs[self.host]
        else:
            raise Exception("No joboutdir found for host " + self.host)

        self.joboutdir = joboutdir
        self.joboutdir_at_host = joboutdir_at_host
        self.ecf_job = self.task.create_ecf_job(joboutdir)
        self.ecf_job_at_host = self.task.create_ecf_job(joboutdir_at_host)
        self.ecf_jobout = self.task.create_ecf_jobout(joboutdir)
        self.ecf_jobout_at_host = self.task.create_ecf_jobout(joboutdir_at_host)

    def check_exceptions(self, submit_exceptions):
        # Check exceptions
        if submit_exceptions is not None:
            for state in submit_exceptions:
                if state == "complete":
                    if "task" in submit_exceptions[state]:
                        for task in submit_exceptions[state]["task"]:
                            if task == self.task.ecf_task:
                                if submit_exceptions[state]["task"][task] == "is_coldstart":
                                    if self.coldstart:
                                        self.complete = "Task " + task + " complete due to cold start"
                    if "family" in submit_exceptions[state]:
                        for family in submit_exceptions[state]["family"]:
                            for ecf_family in self.task.ecf_families:
                                if family == ecf_family:
                                    if submit_exceptions[state]["family"][family] == "is_coldstart":
                                        if self.coldstart:
                                            self.complete = "Family " + family + " complete due to cold start"

    def process_settings(self):
        for key in self.task_settings:
            value = self.task_settings[key]
            # print("key=", key, " value=", value)
            if key == "TRAILER":
                self.trailer.update({key: value})
            else:
                if key == "SUBMIT_TYPE":
                    if value != "":
                        self.submit_type = value
                elif key == "SSH":
                    self.remote_submit_cmd = value
                    self.remote_status_cmd = value
                    self.remote_kill_cmd = value
                elif key == "INTERPRETER":
                    self.interpreter = value
                elif key == "WRAPPER":
                    self.wrapper = value
                elif key == "HOST":
                    self.host = str(value)
                    if self.host != "0" and self.host != "1":
                        raise Exception("Expected a single or dual-host system. HOST=", self.host)
                else:
                    self.header.update({key: value})

    def parse_submission_defs(self):
        ecf_task = self.task.ecf_task
        task_settings = {}
        # print("parse", ecf_task)
        # print(self.submission_defs)
        all_defs = self.submission_defs
        submit_types = all_defs["submit_types"]
        default_submit_type = all_defs["default_submit_type"]
        task_submit_type = None
        for st in submit_types:
            if st in all_defs and "tasks" in all_defs[st]:
                for t in all_defs[st]["tasks"]:
                    if t == ecf_task:
                        task_submit_type = st
        if task_submit_type is None:
            task_submit_type = default_submit_type

        if task_submit_type in all_defs:
            for setting in all_defs[task_submit_type]:
                if setting != "tasks":
                    task_settings.update({setting: all_defs[task_submit_type][setting]})

        ex = "task_exceptions"
        if ex in all_defs:
            for t in all_defs[ex]:
                if t == ecf_task:
                    for setting in all_defs[ex][t]:
                        task_settings.update({setting: all_defs[ex][t][setting]})

        # print(task_settings)
        return task_settings


class SubmitException(Exception):
    def __init__(self, msg, task, task_settings):
        logfile = task.create_submission_log(task_settings.joboutdir_at_host)
        fh = open(logfile, "a")
        fh.write(msg)
        fh.flush()
        fh.close()
        print(msg)
        exit(0)


class KillException(Exception):
    def __init__(self, msg, task, task_settings):
        logfile = task.create_kill_log(task_settings.joboutdir_at_host)
        fh = open(logfile, "a")
        fh.write(msg)
        fh.flush()
        fh.close()
        print(msg)
        exit(0)


class StatusException(Exception):
    def __init__(self, msg, task, task_settings):
        # joboutdir = task.joboutdir
        logfile = task.create_status_log(task_settings.joboutdir_at_host)
        fh = open(logfile, "a")
        fh.write(msg)
        fh.flush()
        fh.close()
        print(msg)
        exit(0)


def get_submission_object(task, task_settings, db_file=None):

    submit_type = task_settings.submit_type
    print(submit_type)
    if submit_type.lower() == "pbs":
        sub = PBSSubmission(task, task_settings, db=db_file)
    elif submit_type.lower() == "slurm":
        sub = SlurmSubmission(task, task_settings, db=db_file)
    elif submit_type.lower() == "grid_engine":
        sub = GridEngineSubmission(task, task_settings, db=db_file)
    elif submit_type == "background":
        sub = BackgroundSubmission(task, task_settings, db=db_file)
    else:
        raise NotImplementedError
    return sub


class SubmissionBaseClass(ABC):
    def __init__(self, task, task_settings, db=None, remote_submit_cmd=None, remote_kill_cmd=None,
                 remote_status_cmd=None):
        self.task = task
        self.task_settings = task_settings
        # self.sub_logfile = sub_logfile
        self.db = db
        self.process = None
        self.job_id = None
        if task.submission_id is not None:
            self.job_id = task.submission_id
        self.submit_cmd = None
        self.kill_job_cmd = None
        self.job_status_cmd = None

        self.remote_submit_cmd = remote_submit_cmd
        self.remote_kill_cmd = remote_kill_cmd
        self.remote_status_cmd = remote_status_cmd

    def update_db(self, job_id):
        if self.db is not None:
            fh = open(self.db, "a")
            fh.write(job_id + "\n")
            fh.close()

    def clear_db(self):
        if self.db is not None:
            if os.path.exists(self.db):
                os.unlink(self.db)

    @abstractmethod
    def set_submit_cmd(self):
        raise NotImplementedError

    @abstractmethod
    def set_jobid(self):
        raise NotImplementedError

    @abstractmethod
    def get_logfile(self):
        raise NotImplementedError

    def submit_job(self):
        print(self.submit_cmd)
        if self.submit_cmd is not None:
            cmd = self.set_remote_cmd(self.submit_cmd, self.remote_submit_cmd)
            logfile = self.get_logfile()
            print(cmd)
            print(logfile)
            if logfile is None:
                subfile = self.task.create_submission_log(self.task_settings.joboutdir_at_host)
                subfile = open(subfile, "w")
                print(cmd)
                process = subprocess.Popen(cmd, stdout=subfile, stderr=subfile, shell=True)
                process.wait()
                subfile.close()
                ret = process.returncode
                if ret != 0:
                    raise RuntimeError("Submit command failed with error code " + str(ret))
            else:
                logfile = open(logfile, "w") 
                self.process = subprocess.Popen(cmd.split(), stdout=logfile, stderr=logfile)
                
            # print(self.submit_cmd)
            self.job_id = self.set_jobid()
            # print(self.job_id)
            if self.db is not None:
                SubmissionBaseClass.update_db(self, self.job_id)

    def kill_job(self):

        if self.kill_job_cmd is not None:
            killfile = self.task.create_kill_log(self.task_settings.joboutdir_at_host)
            # print(self.kill_job_cmd)
            cmd = self.set_remote_cmd(self.kill_job_cmd, self.remote_kill_cmd)
            fh = open(killfile, "w")
            fh.write("Kill job " + self.task_settings.ecf_job_at_host + " with command:\n")
            fh.write(cmd + "\n")
            fh.flush()
            fh.close()

            stdout = open(killfile, "a")
            process = subprocess.Popen(cmd, stdout=stdout, stderr=stdout, shell=True)
            process.wait()
            ret = process.returncode
            if ret != 0:
                raise RuntimeError("Kill command failed with error code " + str(ret))

            logfile = self.task_settings.ecf_jobout_at_host
            if os.path.exists(logfile):
                mode = "a"
            else:
                mode = "w"
            fh = open(logfile, mode)
            fh.write("\n\n*** KILLED BY ECF_kill ****")
            fh.flush()
            fh.close()

    @abstractmethod
    def set_job_status(self,):
        raise NotImplementedError

    def status(self):
        if self.job_id is None:
            StatusException("No job ID was provided!", self.task, self.task_settings)
        try:
            self.set_job_status()
        except Exception as error:
            raise StatusException("Setting of status command failed " + repr(error), self.task, self.task_settings)
        if self.job_status_cmd is None:
            raise StatusException("No status command set for " + self.task_settings.submit_type,
                                  self.task, self.task_settings)
        try:
            self.job_status()
        except Exception as error:
            StatusException("Status command failed " + repr(error), self.task, self.task_settings)

    def job_status(self):
        print(self.job_status_cmd)
        if self.job_status_cmd is not None:
            statusfile = self.task.create_status_log(self.task_settings.joboutdir_at_host)
            cmd = self.set_remote_cmd(self.job_status_cmd, self.remote_status_cmd)
            stdout = open(statusfile, "w")
            process = subprocess.Popen(cmd, stdout=stdout, stderr=stdout, shell=True)
            process.wait()
            ret = process.returncode
            if ret != 0:
                raise RuntimeError("Status command failed with error code " + str(ret))

    def kill(self):
        if self.job_id is None:
            raise KillException("No job ID was provided!", self.task, self.task_settings)
        try:
            self.set_kill_cmd()
        except Exception as error:
            raise KillException("Setting of kill command failed " + repr(error), self.task, self.task_settings)
        if self.kill_job_cmd is None:
            raise KillException("No kill command set for " + self.task_settings.submit_type, self.task,
                                self.task_settings)
        try:
            self.kill_job()
        except Exception as error:
            raise KillException("Kill failed " + repr(error), self.task, self.task_settings)

    @abstractmethod
    def set_kill_cmd(self):
        raise NotImplementedError

    @staticmethod
    def set_remote_cmd(cmd, remote_cmd):
        if remote_cmd is not None:
            cmd = remote_cmd + " \"" + str(cmd) + "\""
        return cmd

    @abstractmethod
    def set_output(self):
        raise NotImplementedError

    @abstractmethod
    def set_job_name(self):
        raise NotImplementedError


class BackgroundSubmission(SubmissionBaseClass):
    def __init__(self, task, task_settings, db=None):
        SubmissionBaseClass.__init__(self, task, task_settings, db=db)

    def set_submit_cmd(self):
        ecf_job = self.task_settings.ecf_job_at_host
        self.submit_cmd = self.set_remote_cmd(ecf_job, self.remote_submit_cmd)

    def set_jobid(self):
        return str(self.process.pid)

    def get_logfile(self):
        ecf_jobout = self.task_settings.ecf_jobout_at_host
        return ecf_jobout

    def set_kill_cmd(self):
        if self.job_id is not None:
            cmd = "kill -9 " + str(self.job_id)
            self.kill_job_cmd = self.set_remote_cmd(cmd, self.remote_kill_cmd)

    def set_job_status(self):
        print("set_job_status ", self.job_id)
        print(self.job_status_cmd)
        if self.job_id is not None:
            self.job_status_cmd = "ps -auxq " + str(self.job_id)
        print(self.job_status_cmd)

    def set_output(self):
        string = "# Background jobs use standard output/error\n"
        return string

    def set_job_name(self):
        string = "# Background jobs get job name from process name\n"
        return string


class BatchSubmission(SubmissionBaseClass):
    def __init__(self, task, task_settings, db=None, sub=None, stat=None, kill=None, prefix="#"):
        SubmissionBaseClass.__init__(self, task, task_settings, db=db)
        name = self.task.ecf_name.split("/")
        self.batch_sub = sub
        self.batch_stat = stat
        self.batch_kill = kill
        self.batch_prefix = prefix
        self.name = name[-1]

    def set_submit_cmd(self, remote_cmd=None):
        cmd = self.batch_sub + " " + self.task_settings.ecf_job_at_host
        cmd = self.set_remote_cmd(cmd, self.task_settings.remote_submit_cmd)
        self.submit_cmd = cmd

    def set_jobid(self):
        raise NotImplementedError

    def get_logfile(self):
        return None

    def set_kill_cmd(self):
        if self.job_id is not None:
            self.kill_job_cmd = self.batch_kill + " " + str(self.job_id)

    def set_job_status(self):
        if self.job_id is not None:
            self.job_status_cmd = self.batch_stat + " " + str(self.job_id)

    def set_output(self):
        logfile = self.task_settings.ecf_jobout_at_host
        string = self.batch_prefix + " -o " + logfile + "\n"
        string += self.batch_prefix + " -e " + logfile + "\n"
        string += self.batch_prefix + " -j oe\n"
        return string

    def set_job_name(self):
        string = self.batch_prefix + " -N " + self.name + "\n"
        return string


class PBSSubmission(BatchSubmission):
    def __init__(self, task, task_settings, sub="qsub", stat="qacct -j", kill="qdel", prefix="#PBS", db=None):
        BatchSubmission.__init__(self, task, task_settings, db=db, sub=sub, stat=stat, kill=kill, prefix=prefix)
        self.name = self.name[0:15]

    def set_jobid(self):

        logfile = self.task.create_submission_log(self.task_settings.joboutdir_at_host)
        fh = open(logfile, "r")
        lines = fh.readlines()
        fh.close()

        answer = None
        for line in lines:
            answer = line

        expected_len = 7
        answer = answer.replace("\n", "")
        words = answer.split(" ")
        if len(words) == expected_len:
            # Set job id as the second element in answer
            self.job_id = str(words[2])
        else:
            raise Exception("Expected " + str(expected_len) + " in output. Got " + str(len(words)))

    def set_job_name(self):
        string = self.batch_prefix + " -N " + self.name + "\n"
        return string


class SlurmSubmission(BatchSubmission):
    def __init__(self, task, task_settings, sub="sbatch", stat="sacct -j", kill="scancel", prefix="#SBATCH", db=None):
        BatchSubmission.__init__(self, task, task_settings, db=db, sub=sub, stat=stat, kill=kill, prefix=prefix)
        name = self.task.ecf_name.split("/")
        self.name = name[-1]

    def set_output(self):
        logfile = self.task_settings.ecf_jobout_at_host
        string = self.batch_prefix + " -o " + logfile + "\n"
        string += self.batch_prefix + " -e " + logfile 
        return string

    def set_jobid(self):

        logfile = self.task.create_submission_log(self.task_settings.joboutdir_at_host)
        fh = open(logfile, "r")
        lines = fh.readlines()
        fh.close()

        answer = None
        for line in lines:
            answer = line

        if answer is None:
            raise Exception("No answer found " + str(lines) + " " + logfile)

        expected_len = 4
        answer = answer.replace("\n", "")
        words = answer.split(" ")
        if len(words) == expected_len:
            # Set job id as the second element in answer
            self.job_id = str(words[-1])
        else:
            raise Exception("Expected " + str(expected_len) + " in output. Got " + str(len(words)))

    def set_job_name(self):
        string = self.batch_prefix + " -J " + self.name + "\n"
        return string


class GridEngineSubmission(BatchSubmission):
    def __init__(self, task, task_settings, db=None, sub="qsub", stat="qacct -j", kill="qdel", prefix="#$"):
        BatchSubmission.__init__(self, task, task_settings, db=db, sub=sub, stat=stat, kill=kill, prefix=prefix)
        name = self.task.ecf_name.split("/")
        self.name = name[-1]

    def set_output(self):
        logfile = self.task_settings.ecf_jobout_at_host
        string = self.batch_prefix + " -o " + logfile + "\n"
        string += self.batch_prefix + " -e " + logfile
        return string

    def set_jobid(self):

        # Your job XXXXXX ("name") has been submitted
        logfile = self.task.create_submission_log(self.task_settings.joboutdir_at_host)
        fh = open(logfile, "r")
        lines = fh.readlines()
        fh.close()

        answer = None
        for line in lines:
            answer = line

        expected_len = 7
        answer = answer.replace("\n", "")
        words = answer.split(" ")
        if len(words) == expected_len:
            # Set job id as the second element in answer
            self.job_id = str(words[2])
        else:
            raise Exception("Expected " + str(expected_len) + " in output. Got " + str(len(words)))

    def set_job_name(self):
        string = self.batch_prefix + " -N " + self.name + "\n"
        return string



import unittest
import scheduler
import os
import datetime


class TaskTest(unittest.TestCase):

    def setUp(self):
        progress = {
            "DTG": datetime.datetime(2020, 11, 13, 6),
            "DTGEND": datetime.datetime(2020, 11, 13, 6),
            "DTGBEG": datetime.datetime(2020, 11, 13, 3)
        }
        progress_pp = {"DTGPP": datetime.datetime(2020, 11, 13, 6)}
        self.progress = scheduler.Progress(progress, progress_pp)
        self.stream = None
        self.mbr = None
        self.host = "1"
        self.wrapper = ""
        self.ecf_tryno = 1
        self.ecf_pass = "abcef"
        self.ecf_rid = ""
        self.submission_id = ""
        self.args = None

    @staticmethod
    def _create_exp(wd):
        os.makedirs(wd, exist_ok=True)
        argv = [
            "setup",
            "--wd", wd,
            "-rev", wd,
            "-conf", os.getcwd(),
            "-host", "unittest",
            "--domain_file", "test/settings/conf_proj_test.json"
        ]
        kwargs = scheduler.parse_surfex_script(argv)
        scheduler.surfex_script(**kwargs)

    def test_pgd(self):
        exp_name = "pgd"
        root = os.getcwd()
        wd = "/tmp/host0/hm_wd/" + exp_name
        self._create_exp(wd)
        exp = scheduler.ExpFromFiles(exp_name, wd)
        scheduler.init_run(exp, self.stream)
        wd = "/tmp/host1/scratch/hm_home/" + exp_name + "/lib"

        exp = scheduler.ExpFromFiles(exp_name, wd, host=self.host, stream=self.stream, progress=self.progress)

        ecf_name = "/" + exp_name + "/task"
        task = scheduler.EcflowTask(ecf_name, self.ecf_tryno, self.ecf_pass, self.ecf_rid, self.submission_id)
        scheduler.Pgd(task, exp, host=self.host, mbr=self.mbr, stream=self.stream,
                      args=self.args).run(wrapper=self.wrapper)
        os.chdir(root)


    def test_prep(self):
        exp_name = "prep_task"
        root = os.getcwd()
        wd = "/tmp/host0/hm_wd/" + exp_name
        self._create_exp(wd)
        exp = scheduler.ExpFromFiles(exp_name, wd,)
        scheduler.init_run(exp, self.stream)
        wd = "/tmp/host1/scratch/hm_home/" + exp_name + "/lib"

        exp = scheduler.ExpFromFiles(exp_name, wd, host=self.host, stream=self.stream, progress=self.progress)
        ecf_name = "/" + exp_name + "/task"
        task = scheduler.EcflowTask(ecf_name, self.ecf_tryno, self.ecf_pass, self.ecf_rid, self.submission_id)
        scheduler.Prep(task, exp, host=self.host, mbr=self.mbr, stream=self.stream,
                       args=self.args).run(wrapper=self.wrapper)
        os.chdir(root)

    def test_quality_control_t2m(self):
        exp_name = "quality_control_t2m_task"
        root = os.getcwd()
        wd = "/tmp/host0/hm_wd/" + exp_name
        self._create_exp(wd)
        exp = scheduler.ExpFromFiles(exp_name, wd,)
        scheduler.init_run(exp, self.stream)
        wd = "/tmp/host1/scratch/hm_home/" + exp_name + "/lib"

        exp = scheduler.ExpFromFiles(exp_name, wd, host=self.host, stream=self.stream, progress=self.progress)
        ecf_name = "/" + exp_name + "/task"
        task = scheduler.EcflowTask(ecf_name, self.ecf_tryno, self.ecf_pass, self.ecf_rid, self.submission_id)
        task.family1 = "t2m"
        scheduler.QualityControl(task, exp, host=self.host, mbr=self.mbr, stream=self.stream,
                                 args=self.args).run(wrapper=self.wrapper)
        os.chdir(root)

    def test_quality_control_rh2m(self):
        exp_name = "quality_control_rh2m_task"
        root = os.getcwd()
        wd = "/tmp/host0/hm_wd/" + exp_name
        self._create_exp(wd)
        exp = scheduler.ExpFromFiles(exp_name, wd,)
        scheduler.init_run(exp, self.stream)
        wd = "/tmp/host1/scratch/hm_home/" + exp_name + "/lib"

        exp = scheduler.ExpFromFiles(exp_name, wd, host=self.host, stream=self.stream, progress=self.progress)
        ecf_name = "/" + exp_name + "/task"
        task = scheduler.EcflowTask(ecf_name, self.ecf_tryno, self.ecf_pass, self.ecf_rid, self.submission_id)
        task.family1 = "rh2m"
        scheduler.QualityControl(task, exp, host=self.host, mbr=self.mbr, stream=self.stream,
                                 args=self.args).run(wrapper=self.wrapper)
        os.chdir(root)

    def test_quality_control_sd(self):
        exp_name = "quality_control_sd_task"
        root = os.getcwd()
        wd = "/tmp/host0/hm_wd/" + exp_name
        self._create_exp(wd)
        exp = scheduler.ExpFromFiles(exp_name, wd,)
        scheduler.init_run(exp, self.stream)
        wd = "/tmp/host1/scratch/hm_home/" + exp_name + "/lib"

        exp = scheduler.ExpFromFiles(exp_name, wd, host=self.host, stream=self.stream, progress=self.progress)
        ecf_name = "/" + exp_name + "/task"
        task = scheduler.EcflowTask(ecf_name, self.ecf_tryno, self.ecf_pass, self.ecf_rid, self.submission_id)
        task.family1 = "sd"
        scheduler.QualityControl(task, exp, host=self.host, mbr=self.mbr, stream=self.stream,
                                 args=self.args).run(wrapper=self.wrapper)
        os.chdir(root)



import surfex
import os
import json
import numpy as np
import yaml
from datetime import timedelta
import shutil
import time


class AbstractTask(object):
    def __init__(self, task, exp, **kwargs):
        self.exp = exp
        self.system_file_paths = exp.system_file_paths
        self.config = self.exp.config
        self.dtg = self.exp.progress.dtg
        self.geo = self.config.get_setting("GEOMETRY#GEO")
        if self.geo is None:
            raise Exception("You must set geo!")
        self.task = task
        self.task_settings = None

        if kwargs is not None and "task_settings" in kwargs:
            self.task_settings = kwargs["task_settings"]

        self.mbr = None
        if "mbr" in kwargs:
            self.mbr = kwargs["mbr"]

        self.stream = None
        if "stream" in kwargs:
            self.stream = kwargs["stream"]

        self.host = "0"
        if kwargs is not None and "host" in kwargs:
            self.host = kwargs["host"]

        args = None
        if "args" in kwargs:
            iargs = kwargs["args"]
            if iargs != "" and iargs is not None:
                args = {}
                iargs = iargs.split(" ")
                for a in iargs:
                    var = str(a).split("=")
                    key = var[0]
                    value = var[1]
                    args.update({key: value})
        self.args = args

        masterodb = False
        lfagmap = self.exp.config.get_setting("SURFEX#IO#LFAGMAP", mbr=self.mbr)
        self.csurf_filetype = self.exp.config.get_setting("SURFEX#IO#CSURF_FILETYPE", mbr=self.mbr)
        self.suffix = surfex.SurfFileTypeExtension(self.csurf_filetype, lfagmap=lfagmap, masterodb=masterodb).suffix

        self.wrk = self.system_file_paths.get_system_path("wrk_dir", mbr=self.mbr, basedtg=self.dtg)
        self.archive = self.system_file_paths.get_system_path("archive_dir", mbr=self.mbr, basedtg=self.dtg)
        os.makedirs(self.archive, exist_ok=True)
        self.extrarch = self.system_file_paths.get_system_path("extrarch_dir", mbr=self.mbr, basedtg=self.dtg)
        os.makedirs(self.extrarch, exist_ok=True)
        self.obsdir = self.system_file_paths.get_system_path("obs_dir", mbr=self.mbr, basedtg=self.dtg)
        os.makedirs(self.obsdir, exist_ok=True)
        self.wdir = str(os.getpid())
        self.wdir = self.wrk + "/" + self.wdir
        print("WDIR=" + self.wdir)
        os.makedirs(self.wdir, exist_ok=True)
        os.chdir(self.wdir)

        hh = self.exp.progress.dtg.strftime("%H")
        self.fcint = self.exp.config.get_fcint(hh, mbr=self.mbr)
        self.fg_dtg = self.dtg - timedelta(hours=self.fcint)
        self.input_path = self.exp.wd + "/nam"

        self.fg_guess_sfx = self.wrk + "/first_guess_sfx"
        self.fc_start_sfx = self.wrk + "/fc_start_sfx"

        self.translation = {
            "t2m": "air_temperature_2m",
            "rh2m": "relative_humidity_2m",
            "sd": "surface_snow_thickness"
        }

    def run(self, **kwargs):
        # self.prepare(**kwargs)
        # Add system variables to arguments
        if self.system_file_paths.system_variables is not None:
            kwargs.update({"system_variables": self.system_file_paths.system_variables})
        self.execute(**kwargs)
        self.postfix(**kwargs)

    def execute(self, **kwargs):
        print("WARNING: Using empty base class execute " + str(kwargs))

    def postfix(self, **kwargs):
        print("Base class postfix " + str(kwargs))
        if self.wrk is not None:
            os.chdir(self.wrk)

        if self.wdir is not None:
            shutil.rmtree(self.wdir)


class SurfexBinaryTask(AbstractTask):
    def __init__(self, task, exp, mode, **kwargs):
        AbstractTask.__init__(self, task, exp, **kwargs)
        self.mode = mode
        self.need_pgd = True
        self.need_prep = True
        self.pgd = False
        self.prep = False
        self.perturbed = False
        self.soda = False
        self.namelist = None

    def execute(self, binary, output, **kwargs):

        rte = os.environ
        wrapper = ""
        if "wrapper" in kwargs:
            wrapper = kwargs["wrapper"]

        if self.mode == "pgd":
            self.pgd = True
            self.need_pgd = False
            self.need_prep = False
            input_data = surfex.PgdInputData(self.exp.config, self.system_file_paths, **kwargs)
        elif self.mode == "prep":
            self.prep = True
            self.need_prep = False
            input_data = surfex.PrepInputData(self.exp.config, self.system_file_paths, **kwargs)
        elif self.mode == "offline":
            input_data = surfex.OfflineInputData(self.exp.config, self.system_file_paths, **kwargs)
        elif self.mode == "soda":
            self.soda = True
            print(kwargs)
            input_data = surfex.SodaInputData(self.exp.config, self.system_file_paths, **kwargs)
        elif self.mode == "perturbed":
            self.perturbed = True
            input_data = surfex.OfflineInputData(self.exp.config, self.system_file_paths, **kwargs)
        else:
            raise NotImplementedError(self.mode + " is not implemented!")

        pgd_file_path = None
        if "pgd_file_path" in kwargs:
            pgd_file_path = kwargs["pgd_file_path"]
        prep_file_path = None
        if "prep_file_path" in kwargs:
            prep_file_path = kwargs["prep_file_path"]

        print_namelist = True
        if "print_namelist" in kwargs:
            print_namelist = kwargs["print_namelist"]

        pert = None
        if "pert" in kwargs:
            pert = kwargs["pert"]

        print("pgd", pgd_file_path)
        print(self.perturbed, pert)

        print("kwargs", kwargs)
        self.namelist = surfex.BaseNamelist(self.mode, self.exp.config, self.input_path, **kwargs)

        print("rte", rte)
        batch = surfex.BatchJob(rte, wrapper=wrapper)

        archive_data = None
        if "archive_data" in kwargs:
            archive_data = kwargs["archive_data"]

        # settings = surfex.ascii2nml(json_settings)
        settings = self.namelist.get_namelist()
        self.geo.update_namelist(settings)

        # Create input
        # my_ecoclimap = surfex.JsonInputDataFromFile(ecoclimap_file)
        filetype = settings["nam_io_offline"]["csurf_filetype"]
        pgdfile = settings["nam_io_offline"]["cpgdfile"]
        prepfile = settings["nam_io_offline"]["cprepfile"]
        surffile = settings["nam_io_offline"]["csurffile"]
        lfagmap = False
        if "LFAGMAP" in settings["NAM_IO_OFFLINE"]:
            lfagmap = settings["NAM_IO_OFFLINE"]["LFAGMAP"]

        if self.need_pgd:
            pgdfile = surfex.file.PGDFile(filetype, pgdfile, self.geo, input_file=pgd_file_path,
                                          lfagmap=lfagmap)

        if self.need_prep:
            prepfile = surfex.PREPFile(filetype, prepfile, self.geo, input_file=prep_file_path,
                                       lfagmap=lfagmap)

        if self.need_prep and self.need_pgd:
            surffile = surfex.SURFFile(filetype, surffile, self.geo, archive_file=output, lfagmap=lfagmap)
        else:
            surffile = None

        if self.perturbed:
            surfex.PerturbedOffline(binary, batch, prepfile, pert, settings, input_data,
                                    pgdfile=pgdfile, surfout=surffile,
                                    archive_data=archive_data,
                                    print_namelist=print_namelist)
        elif self.pgd:
            pgdfile = surfex.file.PGDFile(filetype, pgdfile, self.geo, input_file=pgd_file_path,
                                          archive_file=output, lfagmap=lfagmap)
            # print(input_data.data)
            surfex.SURFEXBinary(binary, batch, pgdfile, settings, input_data,
                                archive_data=archive_data, print_namelist=print_namelist)
        elif self.prep:
            prepfile = surfex.PREPFile(filetype, prepfile, self.geo, archive_file=output,
                                       lfagmap=lfagmap)
            surfex.SURFEXBinary(binary, batch, prepfile, settings, input_data, pgdfile=pgdfile,
                                archive_data=archive_data, print_namelist=print_namelist)
        else:
            surfex.SURFEXBinary(binary, batch, prepfile, settings, input_data, pgdfile=pgdfile,
                                surfout=surffile, archive_data=archive_data,
                                print_namelist=print_namelist)


class Pgd(SurfexBinaryTask):
    def __init__(self, task, exp, **kwargs):
        SurfexBinaryTask.__init__(self, task, exp, "pgd", **kwargs)

    def execute(self, **kwargs):
        pgdfile = self.exp.config.get_setting("SURFEX#IO#CPGDFILE") + self.suffix
        output = self.system_file_paths.get_system_file("pgd_dir", pgdfile)
        xyz = self.exp.config.get_setting("COMPILE#XYZ")
        bindir = self.system_file_paths.get_system_path("bin_dir")
        binary = bindir + "/PGD" + xyz

        force = False
        if "force" in kwargs:
            force = kwargs["force"]

        if not os.path.exists(output) or force:
            SurfexBinaryTask.execute(self, binary, output, **kwargs)
        else:
            print("Output already exists: ", output)


class Prep(SurfexBinaryTask):
    def __init__(self, task, exp, **kwargs):
        SurfexBinaryTask.__init__(self, task, exp, "prep", **kwargs)

    def execute(self, **kwargs):

        print("prep ", kwargs)
        pgdfile = self.exp.config.get_setting("SURFEX#IO#CPGDFILE") + self.suffix
        pgd_file_path = self.system_file_paths.get_system_file("pgd_dir", pgdfile)
        prep_file = self.exp.config.get_setting("INITIAL_CONDITIONS#PREP_INPUT_FILE", check_parsing=False)
        prep_file = surfex.SystemFilePaths.substitute_string(prep_file, **kwargs)
        prep_filetype = self.exp.config.get_setting("INITIAL_CONDITIONS#PREP_INPUT_FILETYPE")
        prep_pgdfile = self.exp.config.get_setting("INITIAL_CONDITIONS#PREP_PGDFILE")
        prep_pgdfiletype = self.exp.config.get_setting("INITIAL_CONDITIONS#PREP_PGDFILETYPE")
        prepfile = self.exp.config.get_setting("SURFEX#IO#CPREPFILE") + self.suffix
        output = self.system_file_paths.get_system_file("prep_dir", prepfile, basedtg=self.dtg)
        xyz = self.exp.config.get_setting("COMPILE#XYZ")
        bindir = self.system_file_paths.get_system_path("bin_dir")
        binary = bindir + "/PREP" + xyz

        force = False
        if "force" in kwargs:
            force = kwargs["force"]

        if not os.path.exists(output) or force:

            kwargs = {
                "pgd_file_path": pgd_file_path,
                "prep_file": prep_file,
                "prep_filetype": str(prep_filetype),
                "prep_pgdfile": prep_pgdfile,
                "prep_pgdfiletype": str(prep_pgdfiletype),
                "dtg": self.dtg
            }

            print(kwargs)
            SurfexBinaryTask.execute(self, binary, output, **kwargs)
        else:
            print("Output already exists: ", output)

        # PREP should prepare for forecast
        if os.path.exists(self.fc_start_sfx):
            os.unlink(self.fc_start_sfx)
        os.symlink(output, self.fc_start_sfx)


class Forecast(SurfexBinaryTask):
    def __init__(self, task, exp, **kwargs):
        SurfexBinaryTask.__init__(self, task, exp, "offline", **kwargs)

    def execute(self, **kwargs):

        pgdfile = self.exp.config.get_setting("SURFEX#IO#CPGDFILE") + self.suffix
        pgd_file_path = self.system_file_paths.get_system_file("pgd_dir", pgdfile)
        xyz = self.exp.config.get_setting("COMPILE#XYZ")
        bindir = self.system_file_paths.get_system_path("bin_dir")
        binary = bindir + "/OFFLINE" + xyz
        forc_zs = self.exp.config.get_setting("FORECAST#FORC_ZS")

        kwargs.update({"pgd_file_path": pgd_file_path})
        kwargs.update({"prep_file_path": self.fc_start_sfx})
        kwargs.update({"forc_zs": forc_zs})
        output = self.archive + "/" + self.exp.config.get_setting("SURFEX#IO#CSURFFILE") + self.suffix

        # Forcing dir
        self.system_file_paths.add_system_file_path("forcing_dir", self.system_file_paths.get_system_path(
            "forcing_dir", mbr=self.mbr, basedtg=self.dtg))
        force = False
        if "force" in kwargs:
            force = kwargs["force"]
            del(kwargs["force"])

        if not os.path.exists(output) or force:
            SurfexBinaryTask.execute(self, binary, output, **kwargs)
        else:
            print("Output already exists: ", output)


class PerturbedRun(SurfexBinaryTask):
    def __init__(self, task, exp, **kwargs):
        SurfexBinaryTask.__init__(self, task, exp, "perturbed", **kwargs)
        self.pert = self.args["pert"]

    def execute(self, **kwargs):
        kwargs.update({"pert": self.pert})

        pgdfile = self.exp.config.get_setting("SURFEX#IO#CPGDFILE") + self.suffix
        pgd_file_path = self.system_file_paths.get_system_file("pgd_dir", pgdfile)
        prepfile = self.exp.config.get_setting("SURFEX#IO#CPREPFILE") + self.suffix
        prep_file_path = self.system_file_paths.get_system_file("prep_dir", prepfile, mbr=self.mbr, basedtg=self.fg_dtg)
        xyz = self.exp.config.get_setting("COMPILE#XYZ")
        bindir = self.system_file_paths.get_system_path("bin_dir")
        binary = bindir + "/OFFLINE" + xyz
        forc_zs = self.exp.config.get_setting("FORECAST#FORC_ZS")

        kwargs.update({"pgd_file_path": pgd_file_path})
        kwargs.update({"prep_file_path": prep_file_path})
        kwargs.update({"forc_zs": forc_zs})

        # PREP file is previous analysis unless first assimilation cycle
        if self.fg_dtg == self.exp.progress.get_dtgbeg(self.fcint):
            prep_file = self.exp.config.get_setting("SURFEX#IO#CPREPFILE") + self.suffix
        else:
            prep_file = "ANALYSIS" + self.suffix

        kwargs.update({"prep_file_path": self.system_file_paths.get_system_file("prep_dir", prep_file, mbr=self.mbr,
                                                                                basedtg=self.fg_dtg)})

        output = self.archive + "/" + self.exp.config.get_setting("SURFEX#IO#CSURFFILE") + "_PERT" + self.pert + \
                                                                                           self.suffix

        # Forcing dir is for previous cycle
        # TODO If pertubed runs moved to pp it should be a diffenent dtg
        self.system_file_paths.add_system_file_path("forcing_dir", self.system_file_paths.get_system_path(
            "forcing_dir", mbr=self.mbr, basedtg=self.fg_dtg))

        force = False
        if "force" in kwargs:
            force = kwargs["force"]

        if not os.path.exists(output) or force:
            SurfexBinaryTask.execute(self, binary, output, **kwargs)
        else:
            print("Output already exists: ", output)


class Soda(SurfexBinaryTask):
    def __init__(self, task, exp, **kwargs):
        SurfexBinaryTask.__init__(self, task, exp, "soda", **kwargs)

    def execute(self, **kwargs):
        xyz = self.exp.config.get_setting("COMPILE#XYZ")
        bindir = self.system_file_paths.get_system_path("bin_dir")
        binary = bindir + "/SODA" + xyz

        pgdfile = self.exp.config.get_setting("SURFEX#IO#CPGDFILE") + self.suffix
        pgd_file_path = self.system_file_paths.get_system_file("pgd_dir", pgdfile)

        kwargs.update({"pgd_file_path": pgd_file_path})
        kwargs.update({"prep_file_path": self.fg_guess_sfx})
        kwargs.update({"dtg": self.dtg})
        output = self.archive + "/ANALYSIS" + self.suffix
        if self.exp.config.setting_is("SURFEX#ASSIM#SCHEMES#ISBA", "EKF"):
            # TODO If pertubed runs moved to pp it should be a diffenent dtg
            perturbed_run_dir = self.system_file_paths.get_system_path("archive_dir", check_parsing=False)
            self.system_file_paths.add_system_file_path("perturbed_run_dir", perturbed_run_dir,
                                                        mbr=self.mbr, basedtg=self.dtg)
            perturbed_file_pattern = self.exp.config.get_setting("SURFEX#IO#CSURFFILE") + "_PERT@PERT@" + self.suffix
            kwargs.update({"perturbed_file_pattern": perturbed_file_pattern})

        force = False
        if "force" in kwargs:
            force = kwargs["force"]

        if not os.path.exists(output) or force:
            print(kwargs)
            SurfexBinaryTask.execute(self, binary, output, **kwargs)
        else:
            print("Output already exists: ", output)

        # SODA should prepare for forecast
        if os.path.exists(self.fc_start_sfx):
            os.unlink(self.fc_start_sfx)
        os.symlink(output, self.fc_start_sfx)

    # Make sure we don't clean yet
    def postfix(self, **kwargs):
        pass


class PrepareCycle(AbstractTask):
    def __init__(self, task, exp, **kwargs):
        AbstractTask.__init__(self, task, exp, **kwargs)

    def run(self, **kwargs):
        self.execute(**kwargs)

    def execute(self, **kwargs):
        if os.path.exists(self.wrk):
            shutil.rmtree(self.wrk)


class QualityControl(AbstractTask):
    def __init__(self, task, exp, **kwargs):
        AbstractTask.__init__(self, task, exp, **kwargs)
        self.var_name = task.family1

    def execute(self, **kwargs):

        an_time = self.dtg
        # archive_root = self.get_setting("archive_root")
        input_file = self.exp.wd + "/toml/qc_config.json"
        if os.path.exists(input_file):
            settings_var = json.load(open(input_file, "r"))
            settings = settings_var[self.var_name]
            sfx_lib = self.system_file_paths.get_system_path("sfx_exp_lib")
            settings.update({"domain": {"domain_file": sfx_lib + "/domain.json"}})
            fg_file = self.system_file_paths.get_system_file("archive_dir", "raw.nc", basedtg=self.dtg)
            settings.update({
                "firstguess": {
                    "fg_file": fg_file,
                    "fg_var": self.translation[self.var_name]
                }
            })
        else:
            raise FileNotFoundError("Could not find input file " + input_file)

        print(self.obsdir)
        output = self.obsdir + "/qc_" + self.translation[self.var_name] + ".json"
        try:
            tests = self.config.get_setting("OBSERVATIONS#QC#" + self.var_name.upper() + "#TESTS")
        except Exception as e:
            tests = self.config.get_setting("OBSERVATIONS#QC#TESTS")

        indent = 2
        blacklist = {}
        debug = True
        print(surfex.__file__)
        tests = surfex.titan.define_quality_control(tests, settings, an_time, domain_geo=self.geo, debug=debug,
                                                    blacklist=blacklist)

        if "netatmo" in settings["sets"]:
            filepattern = self.config.get_setting("OBSERVATIONS#NETATMO_FILEPATTERN", check_parsing=False)
            settings["sets"]["netatmo"].update({"filepattern": filepattern})
            print(filepattern)
        if "bufr" in settings["sets"]:
            settings["sets"]["bufr"].update({"filepattern": self.obsdir + "/ob@YYYY@@MM@@DD@@HH@"})

        datasources = surfex.obs.get_datasources(an_time, settings["sets"])
        data_set = surfex.TitanDataSet(self.var_name, settings, tests, datasources, an_time, debug=debug)
        data_set.perform_tests()

        data_set.write_output(output, indent=indent)


class OptimalInterpolation(AbstractTask):
    def __init__(self, task, exp, **kwargs):
        AbstractTask.__init__(self, task, exp, **kwargs)
        self.var_name = task.family1

    def execute(self, **kwargs):

        if self.var_name in self.translation:
            var = self.translation[self.var_name]
        else:
            raise Exception

        hlength = 30000
        vlength = 100000
        wlength = 0.5
        max_locations = 20
        elev_gradient = 0
        epsilon = 0.25

        hlength = self.config.get_setting("OBSERVATIONS#OI#" + self.var_name.upper() + "#HLENGTH", default=hlength)
        vlength = self.config.get_setting("OBSERVATIONS#OI#" + self.var_name.upper() + "#VLENGTH", default=vlength)
        wlength = self.config.get_setting("OBSERVATIONS#OI#" + self.var_name.upper() + "#WLENGTH", default=wlength)
        elev_gradient = self.config.get_setting("OBSERVATIONS#OI#" + self.var_name.upper() + "#GRADIENT",
                                                default=elev_gradient)
        max_locations = self.config.get_setting("OBSERVATIONS#OI#" + self.var_name.upper() + "#MAX_LOCATIONS",
                                                default=max_locations)
        epsilon = self.config.get_setting("OBSERVATIONS#OI#" + self.var_name.upper() + "#EPISLON", default=epsilon)
        minvalue = self.config.get_setting("OBSERVATIONS#OI#" + self.var_name.upper() + "#MINVALUE", default=None,
                                           abort=False)
        maxvalue = self.config.get_setting("OBSERVATIONS#OI#" + self.var_name.upper() + "#MAXVALUE", default=None,
                                           abort=False)
        input_file = self.archive + "/raw_" + var + ".nc"
        output_file = self.archive + "/an_" + var + ".nc"

        # Get input fields
        geo, validtime, background, glafs, gelevs = surfex.read_first_guess_netcdf_file(input_file, var)

        an_time = validtime
        # Read OK observations
        obs_file = self.system_file_paths.get_system_file("obs_dir", "qc_" + var + ".json", basedtg=self.dtg)
        observations = surfex.dataset_from_file(an_time, obs_file, qc_flag=0)

        field = surfex.horizontal_oi(geo, background, observations, gelevs=gelevs, glafs=glafs,
                                     hlength=hlength, vlength=vlength, wlength=wlength,
                                     max_locations=max_locations, elev_gradient=elev_gradient,
                                     epsilon=epsilon, minvalue=minvalue, maxvalue=maxvalue)

        if os.path.exists(output_file):
            os.unlink(output_file)
        surfex.write_analysis_netcdf_file(output_file, field, var, validtime, gelevs, glafs, new_file=True, geo=geo)


class Forcing(AbstractTask):
    def __init__(self, task, exp, **kwargs):
        AbstractTask.__init__(self, task, exp, **kwargs)
        self.var_name = task.family1
        # self.binary = self.config.

    def execute(self, **kwargs):

        dtg = self.dtg
        hh = self.dtg.strftime("%H")
        fcint = self.config.get_fcint(hh, mbr=self.mbr)

        user_config = None
        if user_config is not None:
            user_config = yaml.load(open(kwargs["user_config"]))
            kwargs.update({"user_config": user_config})

        # TODO
        forcing_source = "testdata_grib2"

        # TODO2
        # Use cli interface???

        kwargs.update({"geo_out": self.geo})

        global_config = self.exp.wd + "/pysurfex/surfex/cfg/config.yml"
        global_config = yaml.load(open(global_config, "r"))
        kwargs.update({"config": global_config})

        kwargs.update({"dtg_start": dtg.strftime("%Y%m%d%H")})
        kwargs.update({"dtg_stop": (dtg + timedelta(hours=fcint)).strftime("%Y%m%d%H")})

        forcing_dir = self.system_file_paths.get_system_path("forcing_dir", basedtg=self.dtg)
        os.makedirs(forcing_dir, exist_ok=True)
        output = self.system_file_paths.get_system_file("forcing_dir", "FORCING.nc", basedtg=self.dtg)

        kwargs.update({"of": output})
        kwargs.update({"output_format": "netcdf"})

        if forcing_source == "thredds":
            kwargs.update({"input_format": "netcdf"})
            kwargs.update({"pattern": "https://thredds.met.no/thredds/dodsC/meps25epsarchive/" +
                                      "@YYYY@/@MM@/@DD@/meps_det_2_5km_@YYYY@@MM@@DD@T@HH@Z.nc"})
            kwargs.update({"zref": "ml"})
            kwargs.update({"zval": "constant"})
            kwargs.update({"uref": "ml"})
            kwargs.update({"uval": "constant"})
            kwargs.update({"zsoro_converter": "phi2m"})
            kwargs.update({"sca_sw": "constant"})
            kwargs.update({"co2": "constant"})
            kwargs.update({"rain_converter": "totalprec"})
            kwargs.update({"wind_converter": "windspeed"})
            kwargs.update({"wind_dir_converter": "winddir"})
        elif forcing_source == "testdata_grib2":
            data_path = "/tmp/host1/testdata/"
            kwargs.update({"input_format": "grib2"})
            kwargs.update({"pattern": data_path + "/fc@YYYY@@MM@@DD@@HH@+@LLLL@grib2"})
            kwargs.update({"rain_converter": "totalprec"})
            kwargs.update({"zref": "ml"})
            kwargs.update({"zval": "constant"})
            kwargs.update({"uref": "ml"})
            kwargs.update({"uval": "constant"})
            kwargs.update({"zsoro_converter": "phi2m"})
            kwargs.update({"sca_sw": "constant"})
            kwargs.update({"co2": "constant"})
            kwargs.update({"wind_converter": "windspeed"})
            kwargs.update({"wind_dir_converter": "winddir"})
            kwargs.update({"debug": True})

        if os.path.exists(output):
            print("Output already exists: " + output)
        else:
            options, var_objs, att_objs = surfex.forcing.set_forcing_config(**kwargs)
            surfex.forcing.run_time_loop(options, var_objs, att_objs)


class FirstGuess(AbstractTask):
    def __init__(self, task, exp, **kwargs):
        AbstractTask.__init__(self, task, exp, **kwargs)
        self.var_name = task.family1

    def execute(self, **kwargs):

        firstguess = self.exp.config.get_setting("SURFEX#IO#CSURFFILE") + self.suffix
        fg_file = self.system_file_paths.get_system_file("first_guess_dir", firstguess, basedtg=self.fg_dtg,
                                                         validtime=self.dtg)

        if os.path.islink(self.fg_guess_sfx):
            os.unlink(self.fg_guess_sfx)
        os.symlink(fg_file, self.fg_guess_sfx)


class CycleFirstGuess(AbstractTask):
    def __init__(self, task, exp, **kwargs):
        AbstractTask.__init__(self, task, exp, **kwargs)
        self.var_name = task.family1

    def execute(self, **kwargs):

        firstguess = self.exp.config.get_setting("SURFEX#IO#CSURFFILE") + self.suffix
        fg_file = self.system_file_paths.get_system_file("first_guess_dir", firstguess, basedtg=self.fg_dtg,
                                                         validtime=self.dtg)

        if os.path.islink(self.fc_start_sfx):
            os.unlink(self.fc_start_sfx)
        os.symlink(fg_file, self.fc_start_sfx)


class Oi2soda(AbstractTask):
    def __init__(self, task, exp, **kwargs):
        AbstractTask.__init__(self, task, exp, **kwargs)
        self.var_name = task.family1

    def execute(self, **kwargs):

        yy = self.dtg.strftime("%y")
        mm = self.dtg.strftime("%m")
        dd = self.dtg.strftime("%d")
        hh = self.dtg.strftime("%H")
        obfile = "OBSERVATIONS_" + yy + mm + dd + "H" + hh + ".DAT"
        output = self.system_file_paths.get_system_file("obs_dir", obfile, mbr=self.mbr, basedtg=self.dtg,
                                                        check_existence=False)

        t2m = None
        rh2m = None
        sd = None

        an_variables = {"t2m": False, "rh2m": False, "sd": False}
        nnco = self.exp.config.get_setting("SURFEX#ASSIM#OBS#NNCO")
        snow_ass = self.exp.config.get_setting("SURFEX#ASSIM#ISBA#UPDATE_SNOW_CYCLES")
        snow_ass_done = False
        if len(snow_ass) > 0:
            hh = int(self.exp.progress.dtg.strftime("%H"))
            for sn in snow_ass:
                if hh == int(sn):
                    snow_ass_done = True

        for ivar in range(0, len(nnco)):
            if nnco[ivar] == 1:
                if ivar == 0:
                    an_variables.update({"t2m": True})
                elif ivar == 1:
                    an_variables.update({"rh2m": True})
                elif ivar == 4:
                    if snow_ass_done:
                        an_variables.update({"sd": True})

        for var in an_variables:
            if an_variables[var]:
                var_name = self.translation[var]
                if var == "t2m":
                    t2m = {
                        "file": self.archive + "/raw_" + var_name + ".nc",
                        "var": var_name
                    }
                elif var == "rh2m":
                    rh2m = {
                        "file": self.archive + "/raw_" + var_name + ".nc",
                        "var": var_name
                    }
                elif var == "sd":
                    sd = {
                        "file": self.archive + "/raw_" + var_name + ".nc",
                        "var": var_name
                    }

        surfex.oi2soda(self.dtg, t2m=t2m, rh2m=rh2m, sd=sd, output=output)
        # surfex.run_surfex_binary(binary)


class Qc2obsmon(AbstractTask):
    def __init__(self, task, exp, **kwargs):
        AbstractTask.__init__(self, task, exp, **kwargs)
        self.var_name = task.family1

    def execute(self, **kwargs):

        kwargs.update({"dtg": self.dtg})
        outdir = self.extrarch + "/ecma_sfc/" + self.dtg.strftime("%Y%m%d%H") + "/"
        os.makedirs(outdir, exist_ok=True)
        output = outdir + "/ecma.db"
        kwargs.update({"output": output})

        if os.path.exists(output):
            os.unlink(output)
        nnco = self.exp.config.get_setting("SURFEX#ASSIM#OBS#NNCO")
        for ivar in range(0, len(nnco)):
            if nnco[ivar] == 1:
                if ivar == 0:
                    var_in = "t2m"
                elif ivar == 1:
                    var_in = "rh2m"
                elif ivar == 4:
                    var_in = "sd"
                else:
                    raise NotImplementedError(nnco[ivar])

                if var_in != "sd":
                    var_name = self.translation[var_in]
                    kwargs.update({"qc": self.obsdir + "/qc_" + var_name + ".json"})
                    kwargs.update({"fg_file": self.archive + "/raw_" + var_name + ".nc"})
                    kwargs.update({"an_file": self.archive + "/an_" + var_name + ".nc"})
                    kwargs.update({"varname": var_in})
                    kwargs.update({"file_var": var_name})
                    surfex.write_obsmon_sqlite_file(**kwargs)


class FirstGuess4OI(AbstractTask):
    def __init__(self, task, exp, **kwargs):

        AbstractTask.__init__(self, task, exp, **kwargs)
        self.var_name = task.family1

    def execute(self, **kwargs):

        validtime = self.exp.progress.dtg

        extra = ""
        symlink_files = {}
        if self.var_name in self.translation:
            var = self.translation[self.var_name]
            variables = [var]
            extra = "_" + var
            symlink_files.update({self.archive + "/raw.nc":  "raw" + extra + ".nc"})
        else:
            var_in = []
            nnco = self.exp.config.get_setting("SURFEX#ASSIM#OBS#NNCO")

            for ivar in range(0, len(nnco)):
                if nnco[ivar] == 1:
                    if ivar == 0:
                        var_in.append("t2m")
                    elif ivar == 1:
                        var_in.append("rh2m")
                    elif ivar == 4:
                        var_in.append("sd")

            variables = []
            try:
                for var in var_in:
                    var_name = self.translation[var]
                    variables.append(var_name)
                    symlink_files.update({self.archive + "/raw_" + var_name + ".nc":  "raw.nc"})
            except ValueError:
                raise Exception("Variables could not be translated")

        variables = variables + ["altitude", "land_area_fraction"]

        output = self.archive + "/raw" + extra + ".nc"
        cache_time = 3600
        if "cache_time" in kwargs:
            cache_time = kwargs["cache_time"]
        cache = surfex.cache.Cache(True, cache_time)
        if os.path.exists(output):
            print("Output already exists " + output)
        else:
            self.write_file(output, variables, self.geo, validtime, cache=cache)

        # Create symlinks
        for target in symlink_files:
            linkfile = symlink_files[target]
            if os.path.lexists(target):
                os.unlink(target)
            os.symlink(linkfile, target)

    def write_file(self, output, variables, geo, validtime, cache=None):

        fg = None
        for var in variables:
            try:
                identifier = "INITIAL_CONDITIONS#FG4OI#" + var + "#"
                inputfile = self.config.get_setting(identifier + "INPUTFILE", basedtg=self.fg_dtg, validtime=self.dtg)
            except Exception as e:
                identifier = "INITIAL_CONDITIONS#FG4OI#"
                inputfile = self.config.get_setting(identifier + "INPUTFILE", basedtg=self.fg_dtg, validtime=self.dtg)
            try:
                identifier = "INITIAL_CONDITIONS#FG4OI#" + var + "#"
                fileformat = self.config.get_setting(identifier + "FILEFORMAT")
            except Exception as e:
                identifier = "INITIAL_CONDITIONS#FG4OI#"
                fileformat = self.config.get_setting(identifier + "FILEFORMAT")
            try:
                identifier = "INITIAL_CONDITIONS#FG4OI#" + var + "#"
                converter = self.config.get_setting(identifier + "CONVERTER")
            except Exception as e:
                identifier = "INITIAL_CONDITIONS#FG4OI#"
                converter = self.config.get_setting(identifier + "CONVERTER")

            print(inputfile, fileformat, converter)
            config_file = self.exp.wd + "/pysurfex/surfex/cfg/first_guess.yml"
            config = yaml.load(open(config_file, "r"))
            defs = config[fileformat]
            defs.update({"filepattern": inputfile})

            converter_conf = config[var][fileformat]["converter"]
            if converter not in config[var][fileformat]["converter"]:
                raise Exception("No converter " + converter + " definition found in " + config_file + "!")

            print(converter)
            converter = surfex.read.Converter(converter, validtime, defs, converter_conf, fileformat, validtime)
            field = surfex.read.ConvertedInput(geo, var, converter).read_time_step(validtime, cache)
            field = np.reshape(field, [geo.nlons, geo.nlats])

            # Create file
            if fg is None:
                nx = geo.nlons
                ny = geo.nlats
                fg = surfex.create_netcdf_first_guess_template(variables, nx, ny, output)
                fg.variables["time"][:] = float(validtime.strftime("%s"))
                fg.variables["longitude"][:] = np.transpose(geo.lons)
                fg.variables["latitude"][:] = np.transpose(geo.lats)
                fg.variables["x"][:] = [i for i in range(0, nx)]
                fg.variables["y"][:] = [i for i in range(0, ny)]

            if var == "altitude":
                field[field < 0] = 0

            fg.variables[var][:] = np.transpose(field)

        if fg is not None:
            fg.close()


class MakeOfflineBinaries(AbstractTask):
    def __init__(self, task, exp, **kwargs):
        AbstractTask.__init__(self, task, exp, **kwargs)
        self.var_name = task.family1

    def execute(self, **kwargs):

        print("Not implemented yet")


class LogProgress(AbstractTask):
    def __init__(self, task, exp, **kwargs):
        AbstractTask.__init__(self, task, exp, **kwargs)
        self.var_name = task.family1

    def execute(self, **kwargs):

        stream = None
        if "stream" in kwargs:
            stream = kwargs["stream"]

        progress_file = self.exp.get_file_name(self.exp.wd, "progress", stream=stream, full_path=True)
        progress_pp_file = self.exp.get_file_name(self.exp.wd, "progressPP", stream=stream, full_path=True)

        # Update progress
        cycle = self.exp.progress.dtg.strftime("%H")
        fcint = self.exp.config.get_fcint(cycle, mbr=self.mbr)
        self.exp.progress.increment_progress(fcint, pp=False)
        self.exp.progress.save(progress_file, progress_pp_file, log_pp=False)


class LogProgressPP(AbstractTask):
    def __init__(self, task, exp, **kwargs):
        AbstractTask.__init__(self, task, exp, **kwargs)
        self.var_name = task.family1

    def execute(self, **kwargs):

        progress_file = self.exp.get_file_name(self.exp.wd, "progress", stream=self.stream, full_path=True)
        progress_pp_file = self.exp.get_file_name(self.exp.wd, "progressPP", stream=self.stream, full_path=True)

        # Update progress
        cycle = self.exp.progress.dtg.strftime("%H")
        fcint = self.exp.config.get_fcint(cycle, mbr=self.mbr)
        self.exp.progress.increment_progress(fcint, pp=True)
        self.exp.progress.save(progress_file, progress_pp_file, log=False)


class PrepareOiSoilInput(AbstractTask):
    def __init__(self, task, exp, **kwargs):
        AbstractTask.__init__(self, task, exp, **kwargs)

    def execute(self, **kwargs):
        # Create FG
        raise NotImplementedError


class PrepareOiClimate(AbstractTask):
    def __init__(self, task, exp, **kwargs):
        AbstractTask.__init__(self, task, exp, **kwargs)

    def execute(self, **kwargs):
        # Create CLIMATE.dat
        raise NotImplementedError


class PrepareSST(AbstractTask):
    def __init__(self, task, exp, **kwargs):
        AbstractTask.__init__(self, task, exp, **kwargs)

    def execute(self, **kwargs):
        # Create CLIMATE.dat
        raise NotImplementedError


class PrepareLSM(AbstractTask):
    def __init__(self, task, exp, **kwargs):
        AbstractTask.__init__(self, task, exp, **kwargs)

    def execute(self, **kwargs):

        file = self.archive + "/raw_nc"
        output = self.system_file_paths.get_system_file("climdir", "LSM.DAT", check_existence=False)
        fileformat = "netcdf"
        converter = "none"
        kwargs = {
            "var": "land_area_fraction",
            "file":  file,
            "fileformat": fileformat,
            "output": output,
            "dtg": self.dtg,
            "geo": self.geo,
            "converter": converter,
        }
        print(kwargs)
        surfex.lsm_file_assim(**kwargs)


# Two test cases
class UnitTest(AbstractTask):
    def __init__(self, task, exp, **kwargs):
        AbstractTask.__init__(self, task, exp, **kwargs)

    def execute(self, **kwargs):
        os.makedirs("/tmp/host0/job/test_start_and_run/", exist_ok=True)
        fh = open("/tmp/host1/scratch/hm_home/test_start_and_run/unittest_ok", "w")
        fh.write("ok")
        fh.close()


class SleepingBeauty(AbstractTask):
    def __init__(self, task, exp, **kwargs):
        AbstractTask.__init__(self, task, exp, **kwargs)

    def execute(self, **kwargs):
        print("Sleeping beauty...")
        print("Create /tmp/host1/scratch/hm_home/test_start_and_run/SleepingBeauty")
        os.makedirs("/tmp/host0/job/test_start_and_run/", exist_ok=True)
        fh = open("/tmp/host1/scratch/hm_home/test_start_and_run/SleepingBeauty", "w")
        fh.write("SleepingBeauty")
        fh.close()
        for i in range(0, 20):
            print("sleep.... ", i, "\n")
            time.sleep(1)


class SleepingBeauty2(AbstractTask):
    def __init__(self, task, exp, **kwargs):
        AbstractTask.__init__(self, task, exp, **kwargs)

    def execute(self, **kwargs):
        print("Will the real Sleeping Beauty, please wake up! please wake up!")
        print("Create /tmp/host1/scratch/hm_home/test_start_and_run/SleepingBeauty2")
        os.makedirs("/tmp/host0/job/test_start_and_run/", exist_ok=True)
        fh = open("/tmp/host1/scratch/hm_home/test_start_and_run/SleepingBeauty2", "w")
        fh.write("SleepingBeauty")
        fh.close()


class WakeUpCall(AbstractTask):
    def __init__(self, task, exp, **kwargs):
        AbstractTask.__init__(self, task, exp, **kwargs)

    def execute(self, **kwargs):
        print("This job is default suspended and manually submitted!")
        print("Create /tmp/host1/scratch/hm_home/test_start_and_run/test_submit")
        os.makedirs("/tmp/host0/job/test_start_and_run/", exist_ok=True)
        fh = open("/tmp/host1/scratch/hm_home/test_start_and_run/test_submit", "w")
        fh.write("Job was submitted")
        fh.close()



import sys
import surfex
import scheduler
from argparse import ArgumentParser
from datetime import datetime
import json
import toml
import os
import shutil


def parse_submit_cmd(argv, exp=False):
    """Parse the command line input arguments."""
    parser = ArgumentParser("ECF_submit task to ecflow")
    if exp:
        parser.add_argument('-exp', type=str, help="Name of experiment")
        parser.add_argument('-lib', type=str, help="Library path")
        parser.add_argument('-dtg', dest="dtg", type=str, help="DTG", default=None, required=False)
        parser.add_argument('-dtgbeg', dest="dtgbeg", type=str, help="DTGBEG", default=None, required=False)
    else:
        parser.add_argument('-sub', dest="env_submit", type=str, help="File with submission settings",
                            required=False)
        parser.add_argument('-dir', dest="joboutdir", type=str, help="Ecflow JOBOUTDIR", required=False)
        parser.add_argument('-server', dest="env_server", type=str, help="File or with Ecflow server settimgs",
                            required=False)
        parser.add_argument('--log', dest="logfile", type=str, help="Server logfile", required=True)
    parser.add_argument('-ecf_name', type=str, help="Name of ECF Task")
    parser.add_argument('-ecf_tryno', type=str, help="ECF try number")
    parser.add_argument('-ecf_pass', type=str, help="Name of ECF password")

    parser.add_argument('-stream', type=str, nargs="?", help="Stream", required=False, default=None)
    parser.add_argument('-ecf_rid', nargs='?', type=str, default=None, required=False, help="ECF remote id")
    parser.add_argument('-ensmbr', dest="ensmbr", nargs="?", type=int, help="Ensemble member", required=False,
                        default=None)
    parser.add_argument('--db', dest="dbfile", type=str, nargs="?", help="Database", required=False, default=None)
    parser.add_argument('--version', action='version', version=surfex.__version__)

    if len(argv) == 0:
        parser.print_help()
        sys.exit()

    args = parser.parse_args(argv)
    kwargs = {}
    for arg in vars(args):
        kwargs.update({arg: getattr(args, arg)})
    return kwargs


def submit_cmd(**kwargs):
    ecf_name = kwargs["ecf_name"]
    ensmbr = kwargs["ensmbr"]
    ecf_tryno = kwargs["ecf_tryno"]
    ecf_pass = kwargs["ecf_pass"]
    ecf_rid = kwargs["ecf_rid"]
    joboutdir = kwargs["joboutdir"]
    if isinstance(joboutdir, str):
        joboutdir = {"0": joboutdir}
    env_submit = kwargs["env_submit"]
    if isinstance(env_submit, str):
        env_submit = json.load(open(env_submit, "r"))
    env_server = kwargs["env_server"]
    if isinstance(env_server, str):
        env_server = toml.load(open(env_server, "r"))
    env_file = None
    if "env_file" in kwargs:
        env_file = kwargs["env_file"]

    submission_id = None
    stream = None
    if "stream" in kwargs:
        stream = kwargs["stream"]
    dbfile = None
    if "dbfile" in kwargs:
        dbfile = kwargs["dbfile"]
    coldstart = False
    if "coldstart" in kwargs:
        coldstart = kwargs["coldstart"]

    if ecf_rid is not None:
        if ecf_rid == "":
            ecf_rid = os.getpid()
    else:
        ecf_rid = os.getpid()

    dry_run = False
    if "dry_run" in kwargs:
        dry_run = kwargs["dry_run"]

    try:
        task = scheduler.EcflowTask(ecf_name, ecf_tryno, ecf_pass, ecf_rid, submission_id)
        sub = scheduler.EcflowSubmitTask(task, env_submit, env_server, joboutdir, env_file=env_file, ensmbr=ensmbr,
                                         dbfile=dbfile, stream=stream, coldstart=coldstart)
        if not dry_run:
            sub.submit()
    except Exception as e:
        raise e


def submit_cmd_exp(**kwargs):

    exp = kwargs["exp"]
    lib = kwargs["lib"]
    dtg = kwargs["dtg"]
    if isinstance(dtg, str):
        dtg = datetime.strptime(dtg, "%Y%m%d%H")
    dtgbeg = dtg
    if "dtgbeg" in kwargs:
        dtgbeg = kwargs["dtgbeg"]
        if isinstance(dtgbeg, str):
            dtgbeg = datetime.strptime(dtgbeg, "%Y%m%d%H")

    progress = {
        "DTG": dtg,
        "DTGBEG": dtgbeg,
        "DTGEND": dtg
    }
    progress_pp = {
        "DTGPP": dtg
    }
    progress = scheduler.Progress(progress, progress_pp)

    stream = None
    if "stream" in kwargs:
        stream = kwargs["stream"]
    exp = scheduler.ExpFromFiles(exp, lib, stream=stream, progress=progress)
    wd = exp.wd

    submit_exceptions = wd + "/config//submit/submission.json"
    submit_exceptions = json.load(open(submit_exceptions, "r"))
    kwargs.update({"submit_exceptions": submit_exceptions})
    kwargs.update({"env_file": exp.get_file_name(exp.wd, "env", full_path=True)})
    kwargs.update({"env_submit": exp.env_submit})
    kwargs.update({"env_server": exp.server})
    hosts = exp.system.hosts
    joboutdir = {}
    for host in range(0, len(hosts)):
        joboutdir.update({str(host): exp.system.get_var("JOBOUTDIR", str(host))})
    kwargs.update({"joboutdir": joboutdir})
    scheduler.submit_cmd(**kwargs)


def parse_kill_cmd(argv, exp=False):
    """Parse the command line input arguments."""
    parser = ArgumentParser("Kill EcFlow task and handle abort")
    if exp:
        parser.add_argument('-exp', type=str, help="Name of experiment", required=True)
        parser.add_argument('-lib', type=str, help="Library path", required=True)
    else:
        parser.add_argument("-sub", dest='env_submit', type=str, help="File with submission settings", required=True)
        parser.add_argument('-dir', dest="joboutdirs", type=str, help="Ecflow JOBOUTDIR", required=True)
        parser.add_argument("-server", dest='server', type=str, help="File with Ecflow server settimgs", required=True)
        parser.add_argument('--log', dest="logfile", type=str, help="Server logfile", required=True)
    parser.add_argument("-ecf_name", dest='ecf_name', type=str, help="ECF_NAME", required=True)
    parser.add_argument("-ecf_tryno", dest='ecf_tryno', type=str, help="ECF_TRYNO", required=True)
    parser.add_argument("-ecf_pass", dest='ecf_pass', type=str, help="ECF_PASS", required=True)
    parser.add_argument('-ecf_rid', dest='ecf_rid', type=str, help="ECF_RID", required=False, nargs="?", default=None)
    parser.add_argument('-submission_id', type=str, help="SUBMISSION_ID")
    parser.add_argument('--version', action='version', version=surfex.__version__)

    if len(argv) == 0:
        parser.print_help()
        sys.exit()

    args = parser.parse_args(argv)
    kwargs = {}
    for arg in vars(args):
        kwargs.update({arg: getattr(args, arg)})
    return kwargs


def kill_cmd(**kwargs):

    ecf_name = kwargs["ecf_name"]
    ecf_tryno = kwargs["ecf_tryno"]
    ecf_pass = kwargs["ecf_pass"]
    ecf_rid = kwargs["ecf_rid"]
    submission_id = ""
    if "submission_id" in kwargs:
        submission_id = kwargs["submission_id"]
    if submission_id == "":
        submission_id = None
    env_submit = kwargs["env_submit"]
    if isinstance(env_submit, str):
        env_submit = json.load(open(env_submit, "r"))
    jobout_dirs = kwargs["joboutdirs"]
    if isinstance(jobout_dirs, str):
        jobout_dirs = {"0": jobout_dirs}

    server = kwargs["server"]
    # If a server environment file, create a server
    if isinstance(server, str):
        logfile = kwargs["logfile"]
        server = scheduler.EcflowServerFromFile(server, logfile)

    task = scheduler.EcflowTask(ecf_name, ecf_tryno, ecf_pass, ecf_rid, submission_id)
    task_settings = scheduler.TaskSettings(task, env_submit, jobout_dirs)
    # print(task.submission_id)
    sub = scheduler.get_submission_object(task, task_settings)
    # print(sub)
    sub.kill()
    server.force_aborted(task)


def kill_cmd_exp(**kwargs):

    exp = kwargs["exp"]
    del kwargs["exp"]
    lib = kwargs["lib"]
    del kwargs["lib"]

    exp = scheduler.ExpFromFiles(exp, lib)
    kwargs.update({"env_submit": exp.env_submit})
    kwargs.update({"server": exp.server})
    hosts = exp.system.hosts
    joboutdir = {}
    for host in range(0, len(hosts)):
        joboutdir.update({str(host): exp.system.get_var("JOBOUTDIR", str(host))})
    kwargs.update({"joboutdirs": joboutdir})
    kill_cmd(**kwargs)


def parse_status_cmd(argv, exp=False):
    """Parse the command line input arguments."""
    parser = ArgumentParser("Status of EcFlow task")
    if exp:
        parser.add_argument('-exp', type=str, help="Name of experiment", required=True)
        parser.add_argument('-lib', type=str, help="Library path", required=True)
    else:
        parser.add_argument("-sub", dest='env_submit', type=str, help="File with submission settings", required=True)
        parser.add_argument('-dir', dest="joboutdirs", type=str, help="Ecflow JOBOUTDIR", required=True)
        parser.add_argument("-server", dest='server', type=str, help="File with Ecflow server settimgs", required=True)
        parser.add_argument('--log', dest="logfile", type=str, help="Server logfile", required=True)
    parser.add_argument('-ecf_name', type=str, help="ECF_NAME", required=True)
    parser.add_argument('-ecf_tryno', type=str, help="ECF_TRYNO", required=True)
    parser.add_argument('-ecf_pass', type=str, help="ECF_PASS", required=True)
    parser.add_argument('-ecf_rid', type=str, help="ECF_RID", required=False, nargs="?", default=None)
    parser.add_argument('-submission_id', type=str, help="SUBMISSION_ID")
    parser.add_argument('--version', action='version', version=surfex.__version__)

    if len(argv) == 0:
        parser.print_help()
        sys.exit()

    args = parser.parse_args(argv)
    kwargs = {}
    for arg in vars(args):
        kwargs.update({arg: getattr(args, arg)})
    return kwargs


def status_cmd(**kwargs):

    ecf_name = kwargs["ecf_name"]
    ecf_tryno = kwargs["ecf_tryno"]
    ecf_pass = kwargs["ecf_pass"]
    ecf_rid = kwargs["ecf_rid"]
    submission_id = kwargs["submission_id"]
    env_submit = kwargs["env_submit"]
    joboutdirs = kwargs["joboutdirs"]

    task = scheduler.EcflowTask(ecf_name, ecf_tryno, ecf_pass, ecf_rid, submission_id)
    task_settings = scheduler.TaskSettings(task, env_submit, joboutdirs)

    sub = scheduler.get_submission_object(task, task_settings)
    sub.status()


def status_cmd_exp(**kwargs):

    exp = kwargs["exp"]
    del kwargs["exp"]
    lib = kwargs["lib"]
    del kwargs["lib"]

    exp = scheduler.ExpFromFiles(exp, lib)
    kwargs.update({"env_submit": exp.env_submit})
    kwargs.update({"server": exp.server})
    hosts = exp.system.hosts
    joboutdir = {}
    for host in range(0, len(hosts)):
        joboutdir.update({str(host): exp.system.get_var("JOBOUTDIR", str(host))})
    kwargs.update({"joboutdirs": joboutdir})
    status_cmd(**kwargs)


def parse_surfex_script(argv):
    """Parse the command line input arguments."""
    parser = ArgumentParser("Surfex offline run script")
    parser.add_argument('action', type=str, help="Action", choices=["setup", "start", "prod", "continue", "testbed",
                                                                    "install", "climate", "co"])
    parser.add_argument('-exp', help="Experiment name", type=str, default=None)
    parser.add_argument('--wd', help="Experiment working directory", type=str, default=None)

    parser.add_argument('-dtg', help="DateTimeGroup (YYYYMMDDHH)", type=str, required=False, default=None)
    parser.add_argument('-dtgend', help="DateTimeGroup (YYYYMMDDHH)", type=str, required=False, default=None)
    parser.add_argument('--suite', type=str, default="surfex", required=False, help="Type of suite definition")
    parser.add_argument('--stream', type=str, default=None, required=False, help="Stream")

    # co
    parser.add_argument("--file", type=str, default=None, required=False, help="File to checkout")

    # Setup variables
    parser.add_argument('-rev', dest="rev", help="Surfex source revison", type=str, required=False,
                        default=None)
    parser.add_argument('-conf', dest="conf", help="Surfex configuration", type=str, required=False,
                        default=None)
    parser.add_argument('-host', dest="host", help="Host label for setup files", type=str, required=False,
                        default=None)
    parser.add_argument('--domain_name', help="Domain name", type=str, required=False, default=None)
    parser.add_argument('--domain_file', help="Domain file", type=str, required=False, default=None)
    parser.add_argument('--config', help="Config", type=str, required=False, default=None)
    parser.add_argument('--version', action='version', version=surfex.__version__)

    if len(argv) == 0:
        parser.print_help()
        sys.exit(1)

    args = parser.parse_args(argv)
    kwargs = {}
    for arg in vars(args):
        kwargs.update({arg: getattr(args, arg)})
    return kwargs


def surfex_script(**kwargs):

    action = kwargs["action"]
    exp = None
    if "exp" in kwargs:
        exp = kwargs["exp"]

    # Setup
    wd = None
    if "wd" in kwargs:
        wd = kwargs["wd"]
    rev = None
    if "rev" in kwargs:
        rev = kwargs["rev"]
    conf = None
    if "conf" in kwargs:
        conf = kwargs["conf"]
    host = kwargs["host"]
    config = kwargs["config"]
    domain_name = None
    if "domain_name" in kwargs:
        domain_name = kwargs["domain_name"]
    domain_file = None
    if "domain_file" in kwargs:
        domain_file = kwargs["domain_file"]

    # Co
    file = kwargs["file"]

    # Others
    dtg = kwargs["dtg"]
    dtgend = kwargs["dtgend"]
    suite = kwargs["suite"]
    stream = kwargs["stream"]

    begin = True
    if "begin" in kwargs:
        begin = kwargs["begin"]

    # Find experiment
    if wd is None:
        wd = os.getcwd()
        print("Setting current working directory as WD: " + wd)
    if exp is None:
        print("Setting EXP from WD:" + wd)
        exp = wd.split("/")[-1]
        print("EXP = " + exp)

    if action == "setup":
        # Copy files to WD from REV
        if rev is None:
            raise Exception("You must set REV")
        if host is None:
            raise Exception("You must set host")
        if conf is None:
            conf = rev

        experiment_is_locked = False
        sfx_exp = scheduler.Exp(exp, wd, rev, conf, experiment_is_locked, configuration=config)
        sfx_exp.setup_files(host)

        exp_domain_file = sfx_exp.get_file_name(wd, "domain", full_path=True)
        # print(domain_file, exp_domain_file)
        if domain_file is None:
            if domain_name is not None:
                domain_json = surfex.set_domain(json.load(open(wd + "/config/domains/Harmonie_domains.json", "r")),
                                                domain_name, hm_mode=True)
                json.dump(domain_json, open(exp_domain_file, "w"), indent=2)
            else:
                if not os.path.exists(exp_domain_file):
                    raise Exception("You must provide a domain!")
        else:
            domain_json = json.load(open(domain_file))
            if os.path.abspath(domain_file) != os.path.abspath(exp_domain_file):
                json.dump(domain_json, open(exp_domain_file, "w"), indent=2)

    elif "action" == "mon":
        # TODO

        raise NotImplementedError
    elif action == "co":
        if file is None:
            raise Exception("Checkout requires a file (--file)")
        scheduler.ExpFromFiles(exp, wd).checkout(file)
    else:
        # Some kind of start
        if action == "start" and dtg is None:
            raise Exception("You must provide -dtg to start a simulation")
        elif action == "climate":
            if dtg is None:
                dtg = "2008061600"
            if suite is not None and suite != "climate":
                raise Exception("Action was climate but you also specified a suite not being climate: " + suite)
            suite = "climate"
        elif action == "testbed":
            if dtg is None:
                dtg = "2008061600"
            if suite is not None and suite != "testbed":
                raise Exception("Action was climate but you also specified a suite not being testbed: " + suite)
            suite = "testbed"
        elif action == "install":
            if dtg is None:
                dtg = "2008061600"
            suite = "Makeup"

        updated_progress = {}
        updated_progress_pp = {}
        if dtg is not None:
            updated_progress.update({"DTG": dtg})
        if dtgend is not None:
            updated_progress.update({"DTGEND": dtgend})
        if action == "start":
            if dtg is None:
                raise Exception("No DTG was provided!")
            updated_progress.update({"DTG": dtg})
            updated_progress.update({"DTGBEG": dtg})
            if dtgend is not None:
                updated_progress.update({"DTGEND": dtgend})
            else:
                updated_progress.update({"DTGEND": dtg})
            updated_progress_pp.update({"DTGPP": dtg})

        progress_file = scheduler.Exp.get_file_name(wd, "progress", stream=stream, full_path=True)
        progress_pp_file = scheduler.Exp.get_file_name(wd, "progressPP", stream=stream, full_path=True)

        if action.lower() == "prod" or action.lower() == "continue":
            progress = scheduler.ProgressFromFile(progress_file, progress_pp_file)
            if dtgend is not None:
                progress.dtgend = datetime.strptime(dtgend, "%Y%m%d%H")
        else:
            progress = scheduler.Progress(updated_progress, updated_progress_pp)

        # Update progress
        progress.save(progress_file, progress_pp_file)

        # Set experiment from files. Should be existing now after setup
        sfx_exp = scheduler.ExpFromFiles(exp, wd)
        system = sfx_exp.system

        # Merge config
        all_merged_settings = surfex.merge_toml_env_from_config_dicts(sfx_exp.config_files)
        merged_config, member_merged_config = surfex.process_merged_settings(all_merged_settings)

        # Create configuration
        config = surfex.Configuration(merged_config, member_merged_config)

        data0 = system.get_var("SFX_EXP_DATA", "0")
        lib0 = system.get_var("SFX_EXP_LIB", "0")
        logfile = data0 + "/ECF.log"

        # Create LIB0 and copy init run if WD != lib0
        if wd.rstrip("/") != lib0.rstrip("/"):
            ecf_init_run = lib0 + "/ecf/InitRun.py"
            dirname = os.path.dirname(ecf_init_run)
            # print(dirname)
            dirs = dirname.split("/")
            # print(dirs)
            if len(dirs) > 1:
                p = "/"
                for d in dirs[1:]:
                    p = p + d
                    # print(p)
                    os.makedirs(p, exist_ok=True)
                    p = p + "/"
            shutil.copy2(wd + "/ecf/InitRun.py", ecf_init_run)

        # Create the scheduler
        env_server = sfx_exp.wd + "/Env_server"
        my_scheduler = scheduler.EcflowServerFromFile(env_server, logfile)

        print(scheduler.__file__)
        # Create and start the suite
        def_file = data0 + "/" + suite + ".def"

        my_scheduler.start_exp(config, sfx_exp, suite, def_file, begin=begin)



import unittest
import scheduler
import os
import time
import re


class EcflowTest(unittest.TestCase):

    @staticmethod
    def _create_exp(wd):
        os.makedirs(wd, exist_ok=True)
        argv = [
            "setup",
            "--wd", wd,
            "-rev", wd,
            "-conf", os.getcwd(),
            "-host", "unittest",
            "--domain_file", "test/settings/conf_proj_test.json"
        ]
        kwargs = scheduler.parse_surfex_script(argv)
        scheduler.surfex_script(**kwargs)

    def test_start_and_run(self):
        exp = "test_start_and_run"
        wd = "/tmp/host0/hm_wd/" + exp
        self._create_exp(wd)

        # Dry submit
        argv = [
            "-exp", exp,
            "-lib", wd,
            "-ecf_name", exp + "/dry_submit",
            "-ecf_pass", "12345",
            "-ecf_tryno", "1"
        ]
        print(argv)
        kwargs = scheduler.parse_submit_cmd(argv, exp=True)
        kwargs.update({"dry_run": True})
        scheduler.submit_cmd_exp(**kwargs)

        argv = [
            "start",
            "--wd", wd,
            "-dtg", "2020111303",
            "-dtgend", "2020111306",
            "--suite", "unittest"
        ]
        kwargs = scheduler.parse_surfex_script(argv)
        scheduler.surfex_script(**kwargs)

        # Test if init run has synced
        test_file = "/tmp/host1/scratch/hm_home/" + exp + "/unittest_ok"
        found = False
        for t in range(0, 15):
            print(t)
            if os.path.exists(test_file):
                found = True
                break
            time.sleep(1)
        if not found:
            raise FileNotFoundError(test_file + " not found!")

        test_file = "/tmp/host1/scratch/hm_home/" + exp + "/SleepingBeauty"
        found = False
        for t in range(0, 15):
            print(t)
            if os.path.exists(test_file):
                found = True
                break
            time.sleep(1)
        if not found:
            raise FileNotFoundError(test_file + " not found!")

        ecf_name = None
        ecf_pass = None
        ecf_tryno = None
        ecf_rid = None
        submission_id = None

        # job_file = "/tmp/host1/job/" + exp + "/SleepingBeauty.job1"
        # for line in open(job_file):
        #    for match in re.finditer("^ecf_name = ", line):
        #        print(match)
        #        print(line)
        #        ecf_name = line.split()[2].replace('"', '')
        #    for match in re.finditer("^ecf_pass = ", line):
        #        ecf_pass = line.split()[2].replace('"', '')
        #    for match in re.finditer("^ecf_tryno = ", line):
        #        ecf_tryno = line.split()[2].replace('"', '')
        #    for match in re.finditer("^ecf_rid = ", line):
        #        ecf_rid = line.split()[2].replace('"', '')
        #    for match in re.finditer("^submission_id = ", line):
        #        submission_id = line.split()[2].replace('"', '')

        # Find sleeping beauty information from job file
        sfx_exp = scheduler.ExpFromFiles(exp, wd)
        sfx_exp.server.ecf_client.sync_local()
        defs = sfx_exp.server.ecf_client.get_defs()
        for item in defs.suites:
            for node in item:
                n = node.get_abs_node_path()
                print(n)
                if n == "/" + exp + "/SleepingBeauty":
                    print("There she is")
                    ecf_name = node.find_gen_variable("ECF_NAME").value()
                    ecf_tryno = node.find_gen_variable("ECF_TRYNO").value()
                    ecf_pass = node.find_gen_variable("ECF_PASS").value()
                    ecf_rid = node.find_gen_variable("ECF_RID").value()
                    submission_id = node.find_variable("SUBMISSION_ID").value()
                    print(ecf_name)
                    print(submission_id)

        argv = [
            "-exp", exp,
            "-lib", "/tmp/host1/scratch/hm_home/" + exp + "/lib",
            "-ecf_name", ecf_name,
            "-ecf_pass", ecf_pass,
            "-ecf_tryno", ecf_tryno,
            "-ecf_rid", ecf_rid,
            "-submission_id", submission_id
        ]
        print(argv)
        kwargs = scheduler.parse_status_cmd(argv, exp=True)
        scheduler.status_cmd_exp(**kwargs)

        kwargs = scheduler.parse_kill_cmd(argv, exp=True)
        scheduler.kill_cmd_exp(**kwargs)

        test_file = "/tmp/host1/scratch/hm_home/" + exp + "/SleepingBeauty2"
        found = False
        for t in range(0, 15):
            print(t)
            if os.path.exists(test_file):
                found = True
                break
            time.sleep(1)
        if not found:
            raise FileNotFoundError(test_file + " not found!")

    def test_create_surfex(self):
        exp = "test_create_surfex"
        wd = "/tmp/host0/hm_wd/" + exp
        self._create_exp(wd)

        argv = [
            "start",
            "--wd", wd,
            "-dtg", "2020111303",
            "-dtgend", "2020112319",
        ]
        kwargs = scheduler.parse_surfex_script(argv)
        kwargs.update({"begin": False})
        scheduler.surfex_script(**kwargs)

    # TODO
    def test_parse_surfex_definition(self):
        pass

    @staticmethod
    def test_ecflow_client():

        ecf_host = "localhost"
        ecf_port = (int(os.getuid()) + 1500)
        logfile = "unittest_ECF.log"
        server = scheduler.EcflowServer(ecf_host, ecf_port, logfile)

        def_file = "unittest_test_ecflow.def"
        suite_name = "test_ecflow"
        suite = scheduler.EcflowSuite(suite_name, def_file=def_file)
        fam = scheduler.EcflowSuiteFamily("My_family", suite)
        var = scheduler.EcflowSuiteVariable("ECF_PASS", "FREE")
        scheduler.EcflowSuiteTask("My_task", fam, variables=var)

        suite.save_as_defs()
        server.start_server()
        server.replace(suite_name, def_file)
        server.begin_suite(suite_name)

        ecf_name = "/test_ecflow/My_family/My_task"
        ecf_pass = "FREE"
        ecf_tryno = "1"
        task = scheduler.EcflowTask(ecf_name, ecf_tryno, ecf_pass, ecf_rid=None)

        with scheduler.EcflowClient(server, task) as ci:
            print("Running task ")
            task.submission_id = "test.12345"
            server.update_submission_id(task)
            server.update_log("Hello log")

    def test_ecflow_client_failed(self):

        ecf_host = "localhost"
        ecf_port = (int(os.getuid()) + 1500)
        logfile = "unittest_ECF_failed.log"
        server = scheduler.EcflowServer(ecf_host, ecf_port, logfile)

        def_file = "unittest_test_ecflow_failed.def"
        suite_name = "test_ecflow_failed"
        suite = scheduler.EcflowSuite(suite_name, def_file=def_file)
        fam = scheduler.EcflowSuiteFamily("My_family", suite)
        var = scheduler.EcflowSuiteVariable("ECF_PASS", "FREE")
        scheduler.EcflowSuiteTask("My_task", fam, variables=var)

        suite.save_as_defs()
        server.start_server()
        server.replace(suite_name, def_file)
        server.begin_suite(suite_name)

        ecf_name = "/test_ecflow_failed/My_family/My_task"
        ecf_pass = "FREE"
        ecf_tryno = "1"
        task = scheduler.EcflowTask(ecf_name, ecf_tryno, ecf_pass, ecf_rid=None)

        with self.assertRaises(Exception) as cm:
            with scheduler.EcflowClient(server, task) as ci:
                print("Running task ")
                raise Exception()
